"""An agent to ensure consistency for transformation jobs, tasks and files.

Depending on what is the status of a job and its input and output files different actions are performed.

.. warning:: Before fully enabling this agent make sure that your transformation jobs fulfill the assumptions of the
  agent.  Otherwise it might delete some of your data! Do not set ``EnableFlag`` to True before letting the agent run
  through a few times and read the messages it produces.

.. note::  Use the :ref:`admin_dirac-transformation-recover-data` script for checking individual transformations

The agent takes the following steps

- obtain list of transformation
- get a list of all 'Failed' and 'Done' jobs, jobs with pending requests are ignored.
- get input files for all jobs, get the transformation file status
  associated for the file (Unused, Assigned, MaxReset, Processed),
  check if the input file exists
- get the output files for each job, check if the output files exist
- perform changes for Jobs, Files and Tasks: cleanup incomplete output
  files to obtain consistent state for jobs, tasks, input and output
  files

- Send email about performed actions

Requirements/Assumptions:

- JobParameters:

  - ProductionOutputData: with the semi-colon separated list of expected output files, stored as a Job Parameter
      This parameter needs to be set by the production UploadOutputData tool _before_ uploading files
  - JobName of the form: TransformationID_TaskID obtained as a  JobAttribute
  - InputData from the JobMonitor.getInputData

  - Or Extract that information from the JDL for the job, which must also contain the ProductionOutputData fields

- JobGroup equal to "%08d" % transformationID

.. note::

  Transformations are only treated, if during the last pass changes
  were performed, or the number of Failed and Done jobs has changed.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN DataRecoveryAgent
  :end-before: ##END
  :dedent: 2
  :caption: DataRecoveryAgent options

.. note::

  For the ``TransformationsNoInput`` or ``TransformationsWithInput``
  to take their default value, the options need to be
  removed from the configuration, otherwise no transformations of this type will be treated.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import defaultdict
import time
import itertools

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from DIRAC.TransformationSystem.Client import TransformationFilesStatus
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.TransformationSystem.Utilities.JobInfo import TaskInfoException
from DIRAC.TransformationSystem.Utilities.TransformationInfo import TransformationInfo
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient

__RCSID__ = "$Id$"

AGENT_NAME = "Transformation/DataRecoveryAgent"

ASSIGNEDSTATES = [TransformationFilesStatus.ASSIGNED, TransformationFilesStatus.PROCESSED]


class DataRecoveryAgent(AgentModule):
    """Data Recovery Agent"""

    def __init__(self, *args, **kwargs):
        AgentModule.__init__(self, *args, **kwargs)
        self.name = "DataRecoveryAgent"
        self.enabled = False
        self.getJobInfoFromJDLOnly = False

        self.__getCSOptions()

        # This needs to be both otherwise we cannot account for all cases
        self.jobStatus = [JobStatus.DONE, JobStatus.FAILED]

        self.jobMon = JobMonitoringClient()
        self.fcClient = FileCatalogClient()
        self.tClient = TransformationClient()
        self.reqClient = ReqClient()
        self.diracAPI = Dirac()
        self.inputFilesProcessed = set()
        self.todo = {
            "NoInputFiles": [
                dict(
                    Message="NoInputFiles: OutputExists: Job 'Done'",
                    ShortMessage="NoInputFiles: job 'Done' ",
                    Counter=0,
                    Check=lambda job: job.allFilesExist() and job.status == JobStatus.FAILED,
                    Actions=lambda job, tInfo: [job.setJobDone(tInfo)],
                ),
                dict(
                    Message="NoInputFiles: OutputMissing: Job 'Failed'",
                    ShortMessage="NoInputFiles: job 'Failed' ",
                    Counter=0,
                    Check=lambda job: job.allFilesMissing() and job.status == JobStatus.DONE,
                    Actions=lambda job, tInfo: [job.setJobFailed(tInfo)],
                ),
            ],
            "InputFiles": [  # must always be first!
                dict(
                    Message="One of many Successful: clean others",
                    ShortMessage="Other Tasks --> Keep",
                    Counter=0,
                    Check=lambda job: job.allFilesExist()
                    and job.otherTasks
                    and not set(job.inputFiles).issubset(self.inputFilesProcessed),
                    Actions=lambda job, tInfo: [
                        self.inputFilesProcessed.update(job.inputFiles),
                        job.setJobDone(tInfo),
                        job.setInputProcessed(tInfo),
                    ],
                ),
                dict(
                    Message="Other Task processed Input, no Output: Fail",
                    ShortMessage="Other Tasks --> Fail",
                    Counter=0,
                    Check=lambda job: set(job.inputFiles).issubset(self.inputFilesProcessed)
                    and job.allFilesMissing()
                    and job.status != JobStatus.FAILED,
                    Actions=lambda job, tInfo: [job.setJobFailed(tInfo)],
                ),
                dict(
                    Message="Other Task processed Input: Fail and clean",
                    ShortMessage="Other Tasks --> Cleanup",
                    Counter=0,
                    Check=lambda job: set(job.inputFiles).issubset(self.inputFilesProcessed)
                    and not job.allFilesMissing(),
                    Actions=lambda job, tInfo: [job.setJobFailed(tInfo), job.cleanOutputs(tInfo)],
                ),
                dict(
                    Message="InputFile(s) missing: mark job 'Failed', mark input 'Deleted', clean",
                    ShortMessage="Input Missing --> Job 'Failed, Input 'Deleted', Cleanup",
                    Counter=0,
                    Check=lambda job: job.inputFiles and job.allInputFilesMissing() and not job.allTransFilesDeleted(),
                    Actions=lambda job, tInfo: [
                        job.cleanOutputs(tInfo),
                        job.setJobFailed(tInfo),
                        job.setInputDeleted(tInfo),
                    ],
                ),
                dict(
                    Message="InputFile(s) Deleted, output Exists: mark job 'Failed', clean",
                    ShortMessage="Input Deleted --> Job 'Failed, Cleanup",
                    Counter=0,
                    Check=lambda job: job.inputFiles
                    and job.allInputFilesMissing()
                    and job.allTransFilesDeleted()
                    and not job.allFilesMissing(),
                    Actions=lambda job, tInfo: [job.cleanOutputs(tInfo), job.setJobFailed(tInfo)],
                ),
                # All Output Exists
                dict(
                    Message="Output Exists, job Failed, input not Processed --> Job Done, Input Processed",
                    ShortMessage="Output Exists --> Job Done, Input Processed",
                    Counter=0,
                    Check=lambda job: job.allFilesExist()
                    and not job.otherTasks
                    and job.status == JobStatus.FAILED
                    and not job.allFilesProcessed()
                    and job.allInputFilesExist(),
                    Actions=lambda job, tInfo: [job.setJobDone(tInfo), job.setInputProcessed(tInfo)],
                ),
                dict(
                    Message="Output Exists, job Failed, input Processed --> Job Done",
                    ShortMessage="Output Exists --> Job Done",
                    Counter=0,
                    Check=lambda job: job.allFilesExist()
                    and not job.otherTasks
                    and job.status == JobStatus.FAILED
                    and job.allFilesProcessed()
                    and job.allInputFilesExist(),
                    Actions=lambda job, tInfo: [job.setJobDone(tInfo)],
                ),
                dict(
                    Message="Output Exists, job Done, input not Processed --> Input Processed",
                    ShortMessage="Output Exists --> Input Processed",
                    Counter=0,
                    Check=lambda job: job.allFilesExist()
                    and not job.otherTasks
                    and job.status == JobStatus.DONE
                    and not job.allFilesProcessed()
                    and job.allInputFilesExist(),
                    Actions=lambda job, tInfo: [job.setInputProcessed(tInfo)],
                ),
                # outputmissing
                dict(
                    Message="Output Missing, job Failed, input Assigned, MaxError --> Input MaxReset",
                    ShortMessage="Max ErrorCount --> Input MaxReset",
                    Counter=0,
                    Check=lambda job: job.allFilesMissing()
                    and not job.otherTasks
                    and job.status == JobStatus.FAILED
                    and job.allFilesAssigned()
                    and not set(job.inputFiles).issubset(self.inputFilesProcessed)
                    and job.allInputFilesExist()
                    and job.checkErrorCount(),
                    Actions=lambda job, tInfo: [job.setInputMaxReset(tInfo)],
                ),
                dict(
                    Message="Output Missing, job Failed, input Assigned --> Input Unused",
                    ShortMessage="Output Missing --> Input Unused",
                    Counter=0,
                    Check=lambda job: job.allFilesMissing()
                    and not job.otherTasks
                    and job.status == JobStatus.FAILED
                    and job.allFilesAssigned()
                    and not set(job.inputFiles).issubset(self.inputFilesProcessed)
                    and job.allInputFilesExist(),
                    Actions=lambda job, tInfo: [job.setInputUnused(tInfo)],
                ),
                dict(
                    Message="Output Missing, job Done, input Assigned --> Job Failed, Input Unused",
                    ShortMessage="Output Missing --> Job Failed, Input Unused",
                    Counter=0,
                    Check=lambda job: job.allFilesMissing()
                    and not job.otherTasks
                    and job.status == JobStatus.DONE
                    and job.allFilesAssigned()
                    and not set(job.inputFiles).issubset(self.inputFilesProcessed)
                    and job.allInputFilesExist(),
                    Actions=lambda job, tInfo: [job.setInputUnused(tInfo), job.setJobFailed(tInfo)],
                ),
                # some files missing, needing cleanup. Only checking for
                # assigned, because processed could mean an earlier job was
                # succesful and this one is just the duplicate that needed
                # to be removed! But we check for other tasks earlier, so
                # this should not happen
                dict(
                    Message="Some missing, job Failed, input Assigned --> cleanup, Input 'Unused'",
                    ShortMessage="Output Missing --> Cleanup, Input Unused",
                    Counter=0,
                    Check=lambda job: job.someFilesMissing()
                    and not job.otherTasks
                    and job.status == JobStatus.FAILED
                    and job.allFilesAssigned()
                    and job.allInputFilesExist(),
                    Actions=lambda job, tInfo: [job.cleanOutputs(tInfo), job.setInputUnused(tInfo)],
                ),
                dict(
                    Message="Some missing, job Done, input Assigned --> cleanup, job Failed, Input 'Unused'",
                    ShortMessage="Output Missing --> Cleanup, Job Failed, Input Unused",
                    Counter=0,
                    Check=lambda job: job.someFilesMissing()
                    and not job.otherTasks
                    and job.status == JobStatus.DONE
                    and job.allFilesAssigned()
                    and job.allInputFilesExist(),
                    Actions=lambda job, tInfo: [
                        job.cleanOutputs(tInfo),
                        job.setInputUnused(tInfo),
                        job.setJobFailed(tInfo),
                    ],
                ),
                dict(
                    Message="Some missing, job Done --> job Failed",
                    ShortMessage="Output Missing, Done --> Job Failed",
                    Counter=0,
                    Check=lambda job: not job.allFilesExist() and job.status == "Done",
                    Actions=lambda job, tInfo: [job.setJobFailed(tInfo)],
                ),
                dict(
                    Message="Something Strange",
                    ShortMessage="Strange",
                    Counter=0,
                    Check=lambda job: job.status not in (JobStatus.FAILED, JobStatus.DONE),
                    Actions=lambda job, tInfo: [],
                ),
                # should always be the last one!
                dict(
                    Message="Failed Hard",
                    ShortMessage="Failed Hard",
                    Counter=0,
                    Check=lambda job: False,  # never
                    Actions=lambda job, tInfo: [],
                ),
            ],
        }
        self.jobCache = defaultdict(lambda: (0, 0))
        # Notification options
        self.notesToSend = ""
        self.subject = "DataRecoveryAgent"
        self.startTime = time.time()

        #############################################################################

    def beginExecution(self):
        """Resets defaults after one cycle."""
        self.__getCSOptions()
        return S_OK()

    def __getCSOptions(self):
        """Get agent options from the CS."""
        self.enabled = self.am_getOption("EnableFlag", False)
        self.transformationsToIgnore = self.am_getOption("TransformationsToIgnore", [])
        self.getJobInfoFromJDLOnly = self.am_getOption("JobInfoFromJDLOnly", False)
        self.transformationStatus = self.am_getOption("TransformationStatus", ["Active", "Completing"])
        ops = Operations()
        extendableTTypes = set(ops.getValue("Transformations/ExtendableTransfTypes", ["MCSimulation"]))
        dataProcessing = set(ops.getValue("Transformations/DataProcessing", []))
        self.transNoInput = self.am_getOption("TransformationsNoInput", list(extendableTTypes))
        self.transWithInput = self.am_getOption("TransformationsWithInput", list(dataProcessing - extendableTTypes))
        self.transformationTypes = self.transWithInput + self.transNoInput
        self.log.notice("Will treat transformations without input files", self.transNoInput)
        self.log.notice("Will treat transformations with input files", self.transWithInput)
        self.addressTo = self.am_getOption("MailTo", [])
        self.addressFrom = self.am_getOption("MailFrom", "")
        self.printEveryNJobs = self.am_getOption("PrintEvery", 200)

    def execute(self):
        """The main execution method."""
        self.log.notice("Will ignore the following transformations: %s" % self.transformationsToIgnore)
        self.log.notice(" Job Cache: %s " % self.jobCache)
        transformations = self.getEligibleTransformations(self.transformationStatus, self.transformationTypes)
        if not transformations["OK"]:
            self.log.error("Failure to get transformations", transformations["Message"])
            return S_ERROR("Failure to get transformations")
        for transID, transInfoDict in transformations["Value"].items():
            if transID in self.transformationsToIgnore:
                self.log.notice("Ignoring Transformation: %s" % transID)
                continue
            self.__resetCounters()
            self.inputFilesProcessed = set()
            self.log.notice("Running over Transformation: %s" % transID)
            self.treatTransformation(int(transID), transInfoDict)
            self.sendNotification(transID, transInfoDict)

        return S_OK()

    def getEligibleTransformations(self, status, typeList):
        """Select transformations of given status and type."""
        res = self.tClient.getTransformations(condDict={"Status": status, "Type": typeList})
        if not res["OK"]:
            return res
        transformations = {}
        for prod in res["Value"]:
            transID = prod["TransformationID"]
            transformations[str(transID)] = prod
        return S_OK(transformations)

    def treatTransformation(self, transID, transInfoDict):
        """Run this thing for given transformation."""
        tInfo = TransformationInfo(transID, transInfoDict, self.enabled, self.tClient, self.fcClient, self.jobMon)
        jobs, nDone, nFailed = tInfo.getJobs(statusList=self.jobStatus)

        if not jobs:
            self.log.notice("Skipping. No jobs for transformation", str(transID))
            return

        if self.jobCache[transID][0] == nDone and self.jobCache[transID][1] == nFailed:
            self.log.notice("Skipping transformation %s because nothing changed" % transID)
            return

        self.jobCache[transID] = (nDone, nFailed)

        tasksDict = None
        lfnTaskDict = None

        self.startTime = time.time()
        if transInfoDict["Type"] in self.transWithInput:
            self.log.notice("Getting tasks...")
            tasksDict = tInfo.checkTasksStatus()
            lfnTaskDict = dict(
                [(taskDict["LFN"], taskID) for taskID, taskDicts in tasksDict.items() for taskDict in taskDicts]
            )

        self.checkAllJobs(jobs, tInfo, tasksDict, lfnTaskDict)
        self.printSummary()

    def checkJob(self, job, tInfo):
        """Deal with the job."""
        checks = self.todo["NoInputFiles"] if job.tType in self.transNoInput else self.todo["InputFiles"]
        for do in checks:
            self.log.verbose("Testing: ", do["Message"])
            if do["Check"](job):
                do["Counter"] += 1
                self.log.notice(do["Message"])
                self.log.notice(job)
                self.notesToSend += do["Message"] + "\n"
                self.notesToSend += str(job) + "\n"
                do["Actions"](job, tInfo)
                return

    def getLFNStatus(self, jobs):
        """Get all the LFNs for the jobs and get their status."""
        self.log.notice("Collecting LFNs...")
        lfnExistence = {}
        lfnCache = []
        counter = 0
        jobInfoStart = time.time()
        for counter, job in enumerate(jobs.values(), start=1):
            if counter % self.printEveryNJobs == 0:
                self.log.notice(
                    "Getting JobInfo: %d/%d: %3.1fs" % (counter, len(jobs), float(time.time() - jobInfoStart))
                )
            while True:
                try:
                    job.getJobInformation(self.diracAPI, self.jobMon, jdlOnly=self.getJobInfoFromJDLOnly)
                    lfnCache.extend(job.inputFiles)
                    lfnCache.extend(job.outputFiles)
                    break
                except RuntimeError as e:  # try again
                    self.log.error("+++++ Failure for job:", job.jobID)
                    self.log.error("+++++ Exception: ", str(e))

        timeSpent = float(time.time() - jobInfoStart)
        self.log.notice("Getting JobInfo Done: %3.1fs (%3.3fs per job)" % (timeSpent, timeSpent / counter))

        counter = 0
        fileInfoStart = time.time()
        for lfnChunk in breakListIntoChunks(list(lfnCache), 200):
            counter += 200
            if counter % 1000 == 0:
                self.log.notice(
                    "Getting FileInfo: %d/%d: %3.1fs" % (counter, len(lfnCache), float(time.time() - fileInfoStart))
                )
            while True:
                try:
                    reps = self.fcClient.exists(lfnChunk)
                    if not reps["OK"]:
                        self.log.error("Failed to check file existence, try again...", reps["Message"])
                        raise RuntimeError("Try again")
                    statuses = reps["Value"]
                    lfnExistence.update(statuses["Successful"])
                    break
                except RuntimeError:  # try again
                    pass
        self.log.notice("Getting FileInfo Done: %3.1fs" % (float(time.time() - fileInfoStart)))

        return lfnExistence

    def setPendingRequests(self, jobs):
        """Loop over all the jobs and get requests, if any."""
        for jobChunk in breakListIntoChunks(jobs.values(), 1000):
            jobIDs = [job.jobID for job in jobChunk]
            while True:
                result = self.reqClient.readRequestsForJobs(jobIDs)
                if result["OK"]:
                    break
                self.log.error("Failed to read requests", result["Message"])
                # repeat
            for jobID in result["Value"]["Successful"]:
                request = result["Value"]["Successful"][jobID]
                requestID = request.RequestID
                dbStatus = self.reqClient.getRequestStatus(requestID).get("Value", "Unknown")
                for job in jobChunk:
                    if job.jobID == jobID:
                        job.pendingRequest = dbStatus not in ("Done", "Canceled")
                        self.log.notice(
                            "Found %s request for job %d" % ("pending" if job.pendingRequest else "finished", jobID)
                        )
                        break

    def checkAllJobs(self, jobs, tInfo, tasksDict=None, lfnTaskDict=None):
        """run over all jobs and do checks"""
        fileJobDict = defaultdict(list)
        counter = 0
        nJobs = len(jobs)
        self.setPendingRequests(jobs)
        lfnExistence = self.getLFNStatus(jobs)
        self.log.notice("Running over all the jobs")
        jobCheckStart = time.time()
        for counter, job in enumerate(jobs.values()):
            if counter % self.printEveryNJobs == 0:
                self.log.notice("Checking Jobs %d/%d: %3.1fs" % (counter, nJobs, float(time.time() - jobCheckStart)))
            while True:
                try:
                    if job.pendingRequest:
                        self.log.warn("Job has Pending requests:\n%s" % job)
                        break
                    job.checkFileExistence(lfnExistence)
                    if tasksDict and lfnTaskDict:
                        try:
                            job.getTaskInfo(tasksDict, lfnTaskDict, self.transWithInput)
                        except TaskInfoException as e:
                            self.log.error(" Skip Task, due to TaskInfoException: %s" % e)
                            if not job.inputFiles and job.tType in self.transWithInput:
                                self.__failJobHard(job, tInfo)
                            break
                        for inputFile in job.inputFiles:
                            fileJobDict[inputFile].append(job.jobID)
                    self.checkJob(job, tInfo)
                    break  # get out of the while loop
                except RuntimeError as e:
                    self.log.error("+++++ Failure for job: %d " % job.jobID)
                    self.log.error("+++++ Exception: ", str(e))
                    # run these again because of RuntimeError
        self.log.notice("Checking Jobs Done: %d/%d: %3.1fs" % (counter, nJobs, float(time.time() - jobCheckStart)))

    def printSummary(self):
        """print summary of changes"""
        self.log.notice("Summary:")
        for do in itertools.chain.from_iterable(self.todo.values()):
            message = "%s: %s" % (do["ShortMessage"].ljust(56), str(do["Counter"]).rjust(5))
            self.log.notice(message)
            if self.notesToSend:
                self.notesToSend = str(message) + "\n" + self.notesToSend

    def __resetCounters(self):
        """reset counters for modified jobs"""
        for _name, checks in self.todo.items():
            for do in checks:
                do["Counter"] = 0

    def __failJobHard(self, job, tInfo):
        """set job to failed and remove output files if there are any"""
        if job.inputFiles:
            return
        if job.status in ("Failed",) and job.allFilesMissing():
            return
        self.log.notice("Failing job hard %s" % job)
        self.notesToSend += "Failing job %s: no input file?\n" % job.jobID
        self.notesToSend += str(job) + "\n"
        self.todo["InputFiles"][-1]["Counter"] += 1
        job.cleanOutputs(tInfo)
        job.setJobFailed(tInfo)
        # if job.inputFile is not None:
        #   job.setInputDeleted(tInfo)

    def __notOnlyKeepers(self, transType):
        """check of we only have 'Keep' messages

        in this case we do not have to send report email or run again next time

        """
        if transType in self.transNoInput:
            return True

        checks = self.todo["InputFiles"]
        totalCount = 0
        for check in checks[1:]:
            totalCount += check["Counter"]

        return totalCount > 0

    def sendNotification(self, transID, transInfoDict):
        """Send notification email if something was modified for a transformation.

        :param int transID: ID of given transformation
        :param transInfoDict:
        """
        if not self.addressTo or not self.addressFrom or not self.notesToSend:
            return
        if not self.__notOnlyKeepers(transInfoDict["Type"]):
            # purge notes
            self.notesToSend = ""
            return

        # remove from the jobCache because something happened
        self.jobCache.pop(int(transID), None)
        # send the email to recipients
        for address in self.addressTo:
            result = NotificationClient().sendMail(
                address, "%s: %s" % (self.subject, transID), self.notesToSend, self.addressFrom, localAttempt=False
            )
            if not result["OK"]:
                self.log.error("Cannot send notification mail", result["Message"])
        # purge notes
        self.notesToSend = ""



"""Check options for all agents."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import pytest

from DIRAC.tests.Utilities.assertingUtils import AgentOptionsTest
from DIRAC import S_OK

AGENTS = [
    ("DIRAC.AccountingSystem.Agent.NetworkAgent", {"IgnoreOptions": ["MessageQueueURI"]}),
    (
        "DIRAC.ConfigurationSystem.Agent.Bdii2CSAgent",
        {"IgnoreOptions": ["BannedCEs", "BannedSEs", "DryRun", "AlternativeBDIIs", "VO"]},
    ),
    ("DIRAC.ConfigurationSystem.Agent.GOCDB2CSAgent", {"IgnoreOptions": ["Cycles", "DryRun", "UpdatePerfSONARS"]}),
    ("DIRAC.ConfigurationSystem.Agent.RucioSynchronizerAgent", {}),
    ("DIRAC.ConfigurationSystem.Agent.VOMS2CSAgent", {"IgnoreOptions": ["VO"]}),
    ("DIRAC.DataManagementSystem.Agent.FTS3Agent", {}),
    ("DIRAC.FrameworkSystem.Agent.CAUpdateAgent", {}),
    ("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent", {}),
    (
        "DIRAC.FrameworkSystem.Agent.MyProxyRenewalAgent",
        {"IgnoreOptions": ["MinValidity", "ValidityPeriod", "MinimumLifeTime", "RenewedLifeTime"]},
    ),
    ("DIRAC.RequestManagementSystem.Agent.CleanReqDBAgent", {}),
    (
        "DIRAC.RequestManagementSystem.Agent.RequestExecutingAgent",
        {"IgnoreOptions": ["FTSMode", "OperationHandlers"], "SpecialMocks": {"gConfig": S_OK([])}},
    ),
    ("DIRAC.ResourceStatusSystem.Agent.CacheFeederAgent", {}),
    ("DIRAC.ResourceStatusSystem.Agent.ElementInspectorAgent", {}),
    ("DIRAC.ResourceStatusSystem.Agent.EmailAgent", {}),
    ("DIRAC.ResourceStatusSystem.Agent.RucioRSSAgent", {}),
    ("DIRAC.ResourceStatusSystem.Agent.SiteInspectorAgent", {}),
    ("DIRAC.ResourceStatusSystem.Agent.SummarizeLogsAgent", {}),
    ("DIRAC.ResourceStatusSystem.Agent.TokenAgent", {}),
    ("DIRAC.StorageManagementSystem.Agent.RequestFinalizationAgent", {}),
    ("DIRAC.StorageManagementSystem.Agent.RequestPreparationAgent", {}),
    ("DIRAC.StorageManagementSystem.Agent.StageMonitorAgent", {}),
    ("DIRAC.StorageManagementSystem.Agent.StageRequestAgent", {"IgnoreOptions": ["PinLifetime"]}),
    ("DIRAC.TransformationSystem.Agent.DataRecoveryAgent", {}),
    ("DIRAC.TransformationSystem.Agent.InputDataAgent", {"IgnoreOptions": ["DateKey", "TransformationTypes"]}),
    (
        "DIRAC.TransformationSystem.Agent.MCExtensionAgent",
        {"IgnoreOptions": ["TransformationTypes", "TasksPerIteration", "MaxFailureRate", "MaxWaitingJobs"]},
    ),
    (
        "DIRAC.TransformationSystem.Agent.TaskManagerAgentBase",
        {
            "IgnoreOptions": [
                "PluginLocation",
                "BulkSubmission",
                "shifterProxy",
                "ShifterCredentials",
                "maxNumberOfThreads",
            ]
        },
    ),
    (
        "DIRAC.TransformationSystem.Agent.TransformationAgent",
        {
            "IgnoreOptions": [
                "PluginLocation",
                "transformationStatus",
                "MaxFiles",
                "MaxFilesToProcess",
                "TransformationTypes",
                "ReplicaCacheValidity",
                "NoUnusedDelay",
                "maxThreadsInPool",
            ]
        },
    ),
    ("DIRAC.TransformationSystem.Agent.TransformationCleaningAgent", {"IgnoreOptions": ["EnableFlag", "shifterProxy"]}),
    (
        "DIRAC.TransformationSystem.Agent.ValidateOutputDataAgent",
        {"IgnoreOptions": ["TransformationTypes", "DirectoryLocations", "TransfIDMeta"]},
    ),
    # ('DIRAC.TransformationSystem.Agent.RequestTaskAgent', {}),  # not inheriting from AgentModule
    # ('DIRAC.TransformationSystem.Agent.WorkflowTaskAgent', {}),  # not inheriting from AgentModule
    ("DIRAC.WorkloadManagementSystem.Agent.JobAgent", {"IgnoreOptions": ["JobWrapperTemplate", "MinimumTimeLeft"]}),
    ("DIRAC.WorkloadManagementSystem.Agent.JobCleaningAgent", {}),
    (
        "DIRAC.WorkloadManagementSystem.Agent.PilotStatusAgent",
        {"IgnoreOptions": ["PilotAccountingEnabled", "ClearPilotsDelay", "ClearAbortedPilotsDelay"]},
    ),
    ("DIRAC.WorkloadManagementSystem.Agent.StalledJobAgent", {}),
    ("DIRAC.WorkloadManagementSystem.Agent.StatesAccountingAgent", {}),
    (
        "DIRAC.WorkloadManagementSystem.Agent.SiteDirector",
        {"SpecialMocks": {"findGenericPilotCredentials": S_OK(("a", "b"))}},
    ),
]


LOG = logging.getLogger("Test")


@pytest.mark.parametrize("agentPath, options", AGENTS)
def test_AgentOptions(agentPath, options, caplog, mocker):
    """Check that all options in ConfigTemplate are found in the initialize method, including default values."""
    caplog.set_level(logging.DEBUG)
    AgentOptionsTest(agentPath, options, mocker=mocker)



"""
BaseFormatter
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import logging
import sys


class BaseFormatter(logging.Formatter):
    """
    BaseFormatter is used to format log record to create a string representing a DIRAC log message.
    It is based on the Formatter object of the standard 'logging' library.

    The purpose of this Formatter is only to add a new parameter in the Formatter constructor: options.
    Indeed, all the custom Formatter objects of DIRAC will inherit from it to have this attribute because
    it is a dictionary containing DIRAC specific options useful to create new format.
    """

    def __init__(self, fmt=None, datefmt=None):
        """
        Initialize the formatter without using parameters.
        They are then modified in format()

        :param str fmt: log format: "%(asctime)s UTC %(name)s %(levelname)s: %(message)"
        :param str datefmt: date format: "%Y-%m-%d %H:%M:%S"
        """
        super(BaseFormatter, self).__init__()

    def format(self, record):
        """
        Overriding.
        format is the main method of the Formatter object because it is the method which transforms
        a log record into a string.
        The format and the dateformat are hardcoded to return a DIRAC log format

        :param record: the log record containing all the information about the log message: name, level, threadid...
        """
        # Generate DIRAC base format
        fmt = ""
        timeStamp = ""
        contextComponent = ""
        thread = ""
        level = ""

        if record.headerIsShown:
            if record.timeStampIsShown:
                timeStamp = "%(asctime)s UTC "
            if record.contextIsShown:
                contextComponentList = ["%(componentname)s%(customname)s"]

                # The local_name is normally only provided when using a
                # local sub logger relying on
                # :py:class:`DIRAC.FrameworkSystem.private.standardLogging.Logging.Logging.LocalSubLogger`
                if hasattr(record, "local_name"):
                    contextComponentList += ["/%(local_name)s"]

                contextComponentList += [" "]
                contextComponent = "".join(contextComponentList)
            if record.threadIDIsShown:
                thread = "[%(thread)d] "
            level = "%(levelname)s"
            fmt += "%s%s%s%s: " % (timeStamp, contextComponent, thread, level)

        fmt += "%(message)s%(spacer)s%(varmessage)s"

        # Formatter class does not have the same arguments between Python2 and 3
        # To prevent breaking the Formatter during the migration to Python3, we check the version
        if sys.version_info < (3,):
            self._fmt = fmt
        else:
            self._style._fmt = fmt  # pylint: disable=no-member
        self.datefmt = "%Y-%m-%d %H:%M:%S"
        return super(BaseFormatter, self).format(record)



"""
BaseBody serves as a base class for all the BodyPlugin, as well as documentation
reference for that object.
"""

from DIRAC.Core.Utilities.JEncode import JSerializable


class BaseBody(JSerializable):
    """
    Body plugins are meant to add evolved logic into the process of
    turning a TransformationSystem Task into an RMS Request.

    All your BodyPlugins should inherit from this class. Note that ``BaseBody``
    inherit from :py:class:`DIRAC.Core.Utilities.JEncode.JSerializable`, so you
    should follow the guidelines from this class too.
    """

    def taskToRequest(self, taskID, task, transID):
        """
        This is the only method needed by your plugin. Its role
        is to turn the task into a Request.
        Note that this method is called for one task at the time.

        :param taskID: id for the task
        :param task: dict describing the task
        :param transID: id of the transformation

        :return: A request object
        """
        raise NotImplementedError()



########################################################################
# File :   ResourcesDefaults.py
# Author : Ricardo Graciani
########################################################################
"""
Some Helper class to access Default options for Different Resources (CEs, SEs, Catalags,...)
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from diraccfg import CFG
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgResourceSection, cfgPath, cfgInstallPath, cfgPathToList

__RCSID__ = "$Id$"


def defaultSection(resource):
    """
    Build the path for the Defaults section
    """
    return cfgPath(cfgResourceSection, "Defaults", resource)


def getComputingElementDefaults(ceName="", ceType="", cfg=None, currentSectionPath=""):
    """
    Return cfgDefaults with defaults for the given CEs defined either in arguments or in the provided cfg
    """
    cesCfg = CFG()
    if cfg:
        try:
            cesCfg.loadFromFile(cfg)
            cesPath = cfgInstallPath("ComputingElements")
            if cesCfg.isSection(cesPath):
                for section in cfgPathToList(cesPath):
                    cesCfg = cesCfg[section]
        except Exception:
            return CFG()

    # Overwrite the cfg with Command line arguments
    if ceName:
        if not cesCfg.isSection(ceName):
            cesCfg.createNewSection(ceName)
        if currentSectionPath:
            # Add Options from Command Line
            optionsDict = __getExtraOptions(currentSectionPath)
            for name, value in optionsDict.items():
                cesCfg[ceName].setOption(name, value)  # pylint: disable=no-member
        if ceType:
            cesCfg[ceName].setOption("CEType", ceType)  # pylint: disable=no-member

    ceDefaultSection = cfgPath(defaultSection("ComputingElements"))
    # Load Default for the given type from Central configuration is defined
    ceDefaults = __gConfigDefaults(ceDefaultSection)
    for ceName in cesCfg.listSections():
        if "CEType" in cesCfg[ceName]:
            ceType = cesCfg[ceName]["CEType"]
            if ceType in ceDefaults:
                for option in ceDefaults[ceType].listOptions():  # pylint: disable=no-member
                    if option not in cesCfg[ceName]:
                        cesCfg[ceName].setOption(
                            option, ceDefaults[ceType][option]  # pylint: disable=unsubscriptable-object
                        )

    return cesCfg


def __gConfigDefaults(defaultPath):
    """
    Build a cfg from a Default Section
    """
    from DIRAC import gConfig

    cfgDefaults = CFG()
    result = gConfig.getSections(defaultPath)
    if not result["OK"]:
        return cfgDefaults
    for name in result["Value"]:
        typePath = cfgPath(defaultPath, name)
        cfgDefaults.createNewSection(name)
        result = gConfig.getOptionsDict(typePath)
        if result["OK"]:
            optionsDict = result["Value"]
            for option, value in optionsDict.items():
                cfgDefaults[name].setOption(option, value)

    return cfgDefaults


def __getExtraOptions(currentSectionPath):
    from DIRAC import gConfig

    optionsDict = {}
    if not currentSectionPath:
        return optionsDict
    result = gConfig.getOptionsDict(currentSectionPath)
    if not result["OK"]:
        return optionsDict
    print(result)
    return result["Value"]



"""
Test Levels

- shown
- getLevel
- setLevel

"""

__RCSID__ = "$Id$"

import pytest

from DIRAC.FrameworkSystem.private.standardLogging.LogLevels import LogLevels
from DIRAC.FrameworkSystem.private.standardLogging.test.TestLogUtilities import gLogger, gLoggerReset, cleaningLog


def test_getAllPossibleLevels():
    """
    Make sure getAllPossibleLevels returns every existing level
    """
    _, log, sublog = gLoggerReset()
    levels = ["DEBUG", "VERBOSE", "INFO", "WARN", "NOTICE", "ERROR", "ALWAYS", "FATAL"]
    assert sorted(gLogger.getAllPossibleLevels()) == sorted(levels)
    assert sorted(log.getAllPossibleLevels()) == sorted(levels)
    assert sorted(sublog.getAllPossibleLevels()) == sorted(levels)


def test_setLevelInit():
    """
    Test setLevel and getLevel: initialization of gLogger
    """
    _, log, sublog = gLoggerReset()

    # make sure gLogger and its subloggers have a common level when initialized
    # we don't set the level of gLogger
    assert gLogger.getLevel() == "NOTICE"
    assert log.getLevel() == "NOTICE"
    assert sublog.getLevel() == "NOTICE"


def test_setLevelgLogger():
    """
    Test setLevel and getLevel: set gLogger level
    """
    _, log, sublog = gLoggerReset()
    levels = gLogger.getAllPossibleLevels()

    # as log and sublog levels has not been changed, they should inherit from the same level
    for level in levels:
        gLogger.setLevel(level)
        assert gLogger.getLevel() == level.upper()
        assert log.getLevel() == level.upper()
        assert sublog.getLevel() == level.upper()


def test_setLevelLog():
    """
    Test setLevel and getLevel: set log level
    """
    _, log, sublog = gLoggerReset()
    levels = gLogger.getAllPossibleLevels()

    # set gLogger level
    gLogger.setLevel("notice")
    # by changing log level, gLogger should not be affected, subLogger should be
    for level in levels:
        log.setLevel(level)
        assert gLogger.getLevel() == "NOTICE"
        assert log.getLevel() == level.upper()
        assert sublog.getLevel() == level.upper()


def test_setLevelSublog():
    """
    Test setLevel and getLevel: set sublog level
    """
    _, log, sublog = gLoggerReset()
    levels = gLogger.getAllPossibleLevels()

    # set gLogger and log level
    gLogger.setLevel("notice")
    log.setLevel("warn")
    # by changing sublog level, gLogger and log should not be affected
    for level in levels:
        sublog.setLevel(level)
        assert gLogger.getLevel() == "NOTICE"
        assert log.getLevel() == "WARN"
        assert sublog.getLevel() == level.upper()


def test_setLevelStopPropagation():
    """
    Test setLevel and getLevel: set gLogger level while log and sublog have already be set
    """
    _, log, sublog = gLoggerReset()
    levels = gLogger.getAllPossibleLevels()

    # by changing gLogger level, log and sublog should not be affected anymore as they have been changed manually
    log.setLevel("warn")
    sublog.setLevel("verbose")
    for level in levels:
        gLogger.setLevel(level)
        assert gLogger.getLevel() == level.upper()
        assert log.getLevel() == "WARN"
        assert sublog.getLevel() == "VERBOSE"


@pytest.mark.parametrize(
    "loggerLevel, isSuperiorTo, logRecordLevel,",
    [("debug", False, "error"), ("error", True, "debug"), ("info", False, "info")],
)
def test_setLevelShowngLogger(loggerLevel, isSuperiorTo, logRecordLevel):
    """
    Set gLogger level: check whether a log record should be displayed
    """
    capturedBackend, log, sublog = gLoggerReset()
    levels = gLogger.getAllPossibleLevels()

    gLogger.setLevel(loggerLevel)

    # convert level name into its integer value
    logRecordLevelValue = LogLevels.getLevelValue(logRecordLevel)
    res = gLogger._createLogRecord(logRecordLevelValue, "message", "")

    # clean the log to remove unecessary information
    logstring = cleaningLog(capturedBackend.getvalue())

    # if loggerLevel is superior to logRecordLevel then:
    # - log record should not appear
    # - shown should return False as the log doesn't appear
    # - value returned by createLogRecord should be False too
    isLoggerLvlSupToLogRecordLvl = LogLevels.getLevelValue(loggerLevel) > logRecordLevelValue
    assert isLoggerLvlSupToLogRecordLvl == isSuperiorTo

    if isLoggerLvlSupToLogRecordLvl:
        assert not gLogger.shown(logRecordLevel)
        assert not res
        assert logstring == ""
    else:
        assert gLogger.shown(logRecordLevel)
        assert res
        assert logstring == "Framework %s: message\n" % logRecordLevel.upper()
        capturedBackend.truncate(0)
        capturedBackend.seek(0)


@pytest.mark.parametrize(
    "loggerLevel, isSuperiorTo, logRecordLevel,",
    [("debug", False, "error"), ("error", True, "debug"), ("info", False, "info")],
)
def test_setLevelShownLog(loggerLevel, isSuperiorTo, logRecordLevel):
    """
    Set log level: check whether a log record should be displayed
    """
    capturedBackend, log, sublog = gLoggerReset()
    levels = gLogger.getAllPossibleLevels()

    gLogger.setLevel("notice")
    log.setLevel(loggerLevel)

    # convert level name into its integer value
    logRecordLevelValue = LogLevels.getLevelValue(logRecordLevel)
    res = log._createLogRecord(logRecordLevelValue, "message", "")

    # clean the log to remove unecessary information
    logstring = cleaningLog(capturedBackend.getvalue())

    # if loggLevel is superior to logRecordLevel then:
    # - log record should not appear
    # - shown should return False as the log doesn't appear
    # - value returned by createLogRecord should be False too
    isLoggerLvlSupToLogRecordLvl = LogLevels.getLevelValue(loggerLevel) > logRecordLevelValue
    assert isLoggerLvlSupToLogRecordLvl == isSuperiorTo

    if LogLevels.getLevelValue(loggerLevel) > logRecordLevelValue:
        assert not log.shown(logRecordLevel)
        assert not res
        assert logstring == ""
    else:
        assert log.shown(logRecordLevel)
        assert res
        assert logstring == "Framework/log %s: message\n" % logRecordLevel.upper()
        capturedBackend.truncate(0)
        capturedBackend.seek(0)


@pytest.mark.parametrize(
    "loggerLevel, isSuperiorTo, logRecordLevel,",
    [("debug", False, "error"), ("error", True, "debug"), ("info", False, "info")],
)
def test_setLevelShownSubLog(loggerLevel, isSuperiorTo, logRecordLevel):
    """
    Set sublog level: check whether a log record should be displayed
    """
    capturedBackend, log, sublog = gLoggerReset()
    levels = gLogger.getAllPossibleLevels()

    gLogger.setLevel("notice")
    log.setLevel("warn")
    sublog.setLevel(loggerLevel)

    # convert level name into its integer value
    logRecordLevelValue = LogLevels.getLevelValue(logRecordLevel)
    res = sublog._createLogRecord(logRecordLevelValue, "message", "")

    # clean the log to remove unecessary information
    logstring = cleaningLog(capturedBackend.getvalue())

    # if logLevel is superior to logRecordLevel then:
    # - log record should not appear
    # - shown should return False as the log doesn't appear
    # - value returned by createLogRecord should be False too
    isLoggerLvlSupToLogRecordLvl = LogLevels.getLevelValue(loggerLevel) > logRecordLevelValue
    assert isLoggerLvlSupToLogRecordLvl == isSuperiorTo

    if LogLevels.getLevelValue(loggerLevel) > logRecordLevelValue:
        assert not sublog.shown(logRecordLevel)
        assert not res
        assert logstring == ""
    else:
        assert sublog.shown(logRecordLevel)
        assert res
        assert logstring == "Framework/log/sublog %s: message\n" % logRecordLevel.upper()
    capturedBackend.truncate(0)
    capturedBackend.seek(0)



"""
:mod: Pfn

.. module: Pfn

:synopsis: pfn URI (un)parsing

.. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# # imports
import os

# # from DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from six.moves.urllib import parse as urlparse


def pfnunparse(pfnDict, srmSpecific=True):
    """Wrapper for backward compatibility
    Redirect either to the old hand made style of unparsing
    the pfn, which works for srm, or to the standard one
    which seems to work for the rest

    :param srmSpecific: use the srm specific parser (default True)
    """
    if srmSpecific:
        return srm_pfnunparse(pfnDict)
    return default_pfnunparse(pfnDict)


def srm_pfnunparse(pfnDict):
    """
    Create PFN URI from pfnDict

    :param dict pfnDict:
    """
    # # make sure all keys are in
    allDict = dict.fromkeys(["Protocol", "Host", "Port", "WSUrl", "Path", "FileName"], "")
    if not isinstance(pfnDict, dict):
        return S_ERROR("pfnunparse: wrong type for pfnDict argument, expected a dict, got %s" % type(pfnDict))
    allDict.update(pfnDict)
    pfnDict = allDict

    # # c
    # # /a/b/c
    filePath = os.path.normpath("/" + pfnDict["Path"] + "/" + pfnDict["FileName"]).replace("//", "/")

    # # host
    uri = pfnDict["Host"]
    if pfnDict["Host"]:
        if pfnDict["Port"]:
            # host:port
            uri = "%s:%s" % (pfnDict["Host"], pfnDict["Port"])
        if pfnDict["WSUrl"]:
            if "?" in pfnDict["WSUrl"] and "=" in pfnDict["WSUrl"]:
                # host/wsurl
                # host:port/wsurl
                uri = "%s%s" % (uri, pfnDict["WSUrl"])
            else:
                # host/wsurl
                # host:port/wsurl
                uri = "%s%s?=" % (uri, pfnDict["WSUrl"])

    if pfnDict["Protocol"]:
        if uri:
            # proto://host
            # proto://host:port
            # proto://host:port/wsurl
            uri = "%s://%s" % (pfnDict["Protocol"], uri)
        else:
            # proto:
            uri = "%s:" % pfnDict["Protocol"]

    pfn = "%s%s" % (uri, filePath)

    # c
    # /a/b/c
    # proto:/a/b/c
    # proto://host/a/b/c
    # proto://host:port/a/b/c
    # proto://host:port/wsurl/a/b/c
    return S_OK(pfn)


def default_pfnunparse(pfnDict):
    """
    Create PFN URI from pfnDict

    :param dict pfnDict:
    """

    try:
        if not isinstance(pfnDict, dict):
            return S_ERROR("pfnunparse: wrong type for pfnDict argument, expected a dict, got %s" % type(pfnDict))
        allDict = dict.fromkeys(["Protocol", "Host", "Port", "Path", "FileName", "Options"], "")
        allDict.update(pfnDict)

        scheme = allDict["Protocol"]

        netloc = allDict["Host"]
        if allDict["Port"]:
            netloc += ":%s" % allDict["Port"]

        path = os.path.join(allDict["Path"], allDict["FileName"])
        query = allDict["Options"]

        pr = urlparse.ParseResult(scheme=scheme, netloc=netloc, path=path, params="", query=query, fragment="")

        pfn = pr.geturl()

        return S_OK(pfn)

    except Exception as e:  # pylint: disable=broad-except
        errStr = "Pfn.default_pfnunparse: Exception while unparsing pfn: %s" % pfnDict
        gLogger.exception(errStr, lException=e)
        return S_ERROR(errStr)


def pfnparse(pfn, srmSpecific=True):
    """Wrapper for backward compatibility
    Redirect either to the old hand made style of parsing
    the pfn, which works for srm, or to the standard one
    which seems to work for the rest

    :param srmSpecific: use the srm specific parser (default True)
    """
    if srmSpecific:
        return srm_pfnparse(pfn)
    return default_pfnparse(pfn)


def srm_pfnparse(pfn):
    """
    Parse pfn and save all bits of information into dictionary

    :param str pfn: pfn string
    """
    if not pfn:
        return S_ERROR("wrong 'pfn' argument value in function call, expected non-empty string, got %s" % str(pfn))
    pfnDict = dict.fromkeys(["Protocol", "Host", "Port", "WSUrl", "Path", "FileName"], "")
    try:
        if ":" not in pfn:
            # pfn = /a/b/c
            pfnDict["Path"] = os.path.dirname(pfn)
            pfnDict["FileName"] = os.path.basename(pfn)
        else:
            # pfn = protocol:/a/b/c
            # pfn = protocol://host/a/b/c
            # pfn = protocol://host:port/a/b/c
            # pfn = protocol://host:port/wsurl?=/a/b/c
            pfnDict["Protocol"] = pfn[0 : pfn.index(":")]
            # # remove protocol:
            pfn = pfn[len(pfnDict["Protocol"]) :]
            # # remove :// or :
            pfn = pfn[3:] if pfn.startswith("://") else pfn[1:]
            if pfn.startswith("/"):
                # # /a/b/c
                pfnDict["Path"] = os.path.dirname(pfn)
                pfnDict["FileName"] = os.path.basename(pfn)
            else:
                # # host/a/b/c
                # # host:port/a/b/c
                # # host:port/wsurl?=/a/b/c
                if ":" not in pfn:
                    # # host/a/b/c
                    pfnDict["Host"] = pfn[0 : pfn.index("/")]
                    pfn = pfn[len(pfnDict["Host"]) :]
                    pfnDict["Path"] = os.path.dirname(pfn)
                    pfnDict["FileName"] = os.path.basename(pfn)
                else:
                    # # host:port/a/b/c
                    # # host:port/wsurl?=/a/b/c
                    pfnDict["Host"] = pfn[0 : pfn.index(":")]
                    # # port/a/b/c
                    # # port/wsurl?=/a/b/c
                    pfn = pfn[len(pfnDict["Host"]) + 1 :]
                    pfnDict["Port"] = pfn[0 : pfn.index("/")]
                    # # /a/b/c
                    # # /wsurl?=/a/b/c
                    pfn = pfn[len(pfnDict["Port"]) :]
                    WSUrl = pfn.find("?")
                    WSUrlEnd = pfn.find("=")
                    if WSUrl == -1 and WSUrlEnd == -1:
                        # # /a/b/c
                        pfnDict["Path"] = os.path.dirname(pfn)
                        pfnDict["FileName"] = os.path.basename(pfn)
                    else:
                        # # /wsurl?blah=/a/b/c
                        pfnDict["WSUrl"] = pfn[0 : WSUrlEnd + 1]
                        # # /a/b/c
                        pfn = pfn[len(pfnDict["WSUrl"]) :]
                        pfnDict["Path"] = os.path.dirname(pfn)
                        pfnDict["FileName"] = os.path.basename(pfn)
        return S_OK(pfnDict)
    except Exception:  # pylint: disable=broad-except
        errStr = "Pfn.srm_pfnparse: Exception while parsing pfn: " + str(pfn)
        gLogger.exception(errStr)
        return S_ERROR(errStr)


def default_pfnparse(pfn):
    """
      Parse pfn and save all bits of information into dictionary

    :param str pfn: pfn string
    """
    if not pfn:
        return S_ERROR("wrong 'pfn' argument value in function call, expected non-empty string, got %s" % str(pfn))
    pfnDict = dict.fromkeys(["Protocol", "Host", "Port", "WSUrl", "Path", "FileName"], "")
    try:

        parse = urlparse.urlparse(pfn)
        pfnDict["Protocol"] = parse.scheme
        if ":" in parse.netloc:
            pfnDict["Host"], pfnDict["Port"] = parse.netloc.split(":")
        else:
            pfnDict["Host"] = parse.netloc
        pfnDict["Path"] = os.path.dirname(parse.path)
        pfnDict["FileName"] = os.path.basename(parse.path)
        if parse.query:
            pfnDict["Options"] = parse.query
        return S_OK(pfnDict)
    except Exception as e:  # pylint: disable=broad-except
        errStr = "Pfn.default_pfnparse: Exception while parsing pfn: " + str(pfn)
        gLogger.exception(errStr, lException=e)
        return S_ERROR(errStr)



""" A computing element class that uses sudo
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import pwd
import stat
import errno

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler

from DIRAC.Resources.Computing.ComputingElement import ComputingElement

__RCSID__ = "$Id$"


class SudoComputingElement(ComputingElement):

    #############################################################################
    def __init__(self, ceUniqueID):
        """Standard constructor."""
        super(SudoComputingElement, self).__init__(ceUniqueID)

        self.ceType = "Sudo"
        self.submittedJobs = 0
        self.runningJobs = 0

        self.processors = int(self.ceParameters.get("NumberOfProcessors", 1))
        self.ceParameters["MaxTotalJobs"] = 1

    #############################################################################
    def submitJob(self, executableFile, proxy=None, **kwargs):
        """Method to submit job, overridden from super-class.

        :param str executableFile: file to execute via systemCall.
                                   Normally the JobWrapperTemplate when invoked by the JobAgent.
        :param str proxy: the proxy used for running the job (the payload). It will be dumped to a file.
        """
        payloadProxy = ""
        if proxy:
            self.log.verbose("Setting up proxy for payload")
            result = self.writeProxyToFile(proxy)
            if not result["OK"]:
                return result

            payloadProxy = result["Value"]  # payload proxy file location

            if "X509_USER_PROXY" not in os.environ:
                self.log.error("X509_USER_PROXY variable for pilot proxy not found in local environment")
                return S_ERROR(DErrno.EPROXYFIND, "X509_USER_PROXY not found")

            # See if a fixed value has been given
            payloadUsername = self.ceParameters.get("PayloadUser")

            if payloadUsername:
                self.log.info("Payload username %s from PayloadUser in ceParameters" % payloadUsername)
            else:
                # First username in the sequence to use when running payload job
                # If first is pltXXp00 then have pltXXp01, pltXXp02, ...
                try:
                    baseUsername = self.ceParameters.get("BaseUsername")
                    baseCounter = int(baseUsername[-2:])
                    self.log.info("Base username from BaseUsername in ceParameters : %s" % baseUsername)
                except Exception:
                    # Last chance to get PayloadUsername
                    if "USER" not in os.environ:
                        self.log.error('PayloadUser, BaseUsername and os.environ["USER"] are not properly defined')
                        return S_ERROR(errno.EINVAL, "No correct payload username provided")

                    baseUsername = os.environ["USER"] + "00p00"
                    baseCounter = 0
                    self.log.info("Base username from $USER + 00p00 : %s" % baseUsername)

                # Next one in the sequence
                payloadUsername = baseUsername[:-2] + ("%02d" % (baseCounter + self.submittedJobs))
                self.log.info("Payload username set to %s using jobs counter" % payloadUsername)

            try:
                payloadUID = pwd.getpwnam(payloadUsername).pw_uid
                payloadGID = pwd.getpwnam(payloadUsername).pw_gid
            except KeyError:
                error = S_ERROR('User "' + str(payloadUsername) + '" does not exist!')
                return error

            self.log.verbose("Starting process for monitoring payload proxy")
            gThreadScheduler.addPeriodicTask(
                self.proxyCheckPeriod,
                self.monitorProxy,
                taskArgs=(payloadProxy, payloadUsername, payloadUID, payloadGID),
                executions=0,
                elapsedTime=0,
            )

        # Submit job
        self.log.info("Changing permissions of executable (%s) to 0755" % executableFile)
        self.submittedJobs += 1

        try:
            os.chmod(
                os.path.abspath(executableFile),
                stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH,
            )
        except OSError as x:
            self.log.error("Failed to change permissions of executable to 0755 with exception", "\n%s" % (x))

        result = self.sudoExecute(
            os.path.abspath(executableFile), payloadProxy, payloadUsername, payloadUID, payloadGID
        )
        self.runningJobs -= 1
        if not result["OK"]:
            self.log.error("Failed sudoExecute", result)
            return result

        self.log.debug("Sudo CE result OK")

        return S_OK()

    #############################################################################
    def sudoExecute(self, executableFile, payloadProxy, payloadUsername, payloadUID, payloadGID):
        """Run sudo with checking of the exit status code."""
        # We now implement a file giveaway using groups, to avoid any need to sudo to root.
        # Each payload user must have their own group. The pilot user must be a member
        # of all of these groups. This allows the pilot user to set the group of the
        # payloadProxy file to be that of the payload user. The payload user can then
        # read it and make a copy of it (/tmp/x509up_uNNNN) that it owns. Some grid
        # commands check that the proxy is owned by the current user so the copy stage
        # is necessary.

        # 1) Make sure the payload user can read its proxy via its per-user group
        os.chown(payloadProxy, -1, payloadGID)
        os.chmod(payloadProxy, stat.S_IRUSR + stat.S_IWUSR + stat.S_IRGRP)

        # 2) Now create a copy of the proxy owned by the payload user
        result = shellCall(
            0,
            '/usr/bin/sudo -u %s sh -c "cp -f %s /tmp/x509up_u%d ; chmod 0400 /tmp/x509up_u%d"'
            % (payloadUsername, payloadProxy, payloadUID, payloadUID),
            callbackFunction=self.sendOutput,
        )

        # 3) Make sure the current directory is +rwx by the pilot's group
        #    (needed for InstallDIRAC but not for LHCbInstallDIRAC, for example)
        os.chmod(".", os.stat(".").st_mode | stat.S_IRWXG)

        # Run the executable (the wrapper in fact)
        cmd = "/usr/bin/sudo -u %s " % payloadUsername
        cmd += "PATH=$PATH "
        cmd += "DIRACSYSCONFIG=/scratch/%s/pilot.cfg " % os.environ.get("USER", "")
        cmd += "LD_LIBRARY_PATH=$LD_LIBRARY_PATH "
        cmd += "PYTHONPATH=$PYTHONPATH "
        cmd += "X509_CERT_DIR=$X509_CERT_DIR "
        cmd += "X509_USER_PROXY=/tmp/x509up_u%d sh -c '%s'" % (payloadUID, executableFile)
        self.log.info("CE submission command is: %s" % cmd)
        self.runningJobs += 1
        result = shellCall(0, cmd, callbackFunction=self.sendOutput)
        self.runningJobs -= 1
        if not result["OK"]:
            result["Value"] = (0, "", "")
            return result

        resultTuple = result["Value"]
        status = resultTuple[0]
        stdOutput = resultTuple[1]
        stdError = resultTuple[2]
        self.log.info("Status after the sudo execution is %s" % str(status))
        if status > 128:
            error = S_ERROR(status)
            error["Value"] = (status, stdOutput, stdError)
            return error

        return result

    #############################################################################
    def getCEStatus(self):
        """Method to return information on running and pending jobs."""
        result = S_OK()
        result["SubmittedJobs"] = self.submittedJobs
        result["RunningJobs"] = self.runningJobs
        result["WaitingJobs"] = 0
        # processors
        result["AvailableProcessors"] = self.processors
        return result

    #############################################################################
    def monitorProxy(self, payloadProxy, payloadUsername, payloadUID, payloadGID):
        """Monitor the payload proxy and renew as necessary."""
        retVal = self._monitorProxy(payloadProxy)
        if not retVal["OK"]:
            # Failed to renew the proxy, nothing else to be done
            return retVal

        if not retVal["Value"]:
            # No need to renew the proxy, nothing else to be done
            return retVal

        self.log.info("Re-executing sudo to make renewed payload proxy available as before")

        # New version of the proxy file, so we have to do the copy again

        # 1) Make sure the payload user can read its proxy via its per-user group
        os.chown(payloadProxy, -1, payloadGID)
        os.chmod(payloadProxy, stat.S_IRUSR + stat.S_IWUSR + stat.S_IRGRP)

        # 2) Now recreate the copy of the proxy owned by the payload user
        cmd = '/usr/bin/sudo -u %s sh -c "cp -f %s /tmp/x509up_u%d ; chmod 0400 /tmp/x509up_u%d"' % (
            payloadUsername,
            payloadProxy,
            payloadUID,
            payloadUID,
        )
        result = shellCall(0, cmd, callbackFunction=self.sendOutput)
        if not result["OK"]:
            self.log.error("Could not recreate the copy of the proxy", "CMD: %s; %s" % (cmd, result["Message"]))

        return S_OK("Proxy checked")



#!/usr/bin/env python
"""
Remove the given file replica or a list of file replicas from the File Catalog
and from the storage.

Example:
  $ dirac-dms-remove-replicas /formation/user/v/vhamar/Test.txt IBCP-disk
  Successfully removed DIRAC-USER replica of /formation/user/v/vhamar/Test.txt
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC import exit as DIRACExit
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(("LocalFile: Path to local file containing LFNs", "LFN:       Logical File Names"))
    Script.registerArgument(["SE:        Storage element"])

    Script.parseCommandLine()

    from DIRAC.Core.Utilities.List import breakListIntoChunks
    from DIRAC.DataManagementSystem.Client.DataManager import DataManager

    dm = DataManager()
    import os

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    first, storageElementNames = Script.getPositionalArgs(group=True)

    if os.path.exists(first):
        with open(first, "r") as inputFile:
            string = inputFile.read()
        lfns = [lfn.strip() for lfn in string.splitlines()]
        inputFile.close()
    else:
        lfns = [first]

    for lfnList in breakListIntoChunks(sorted(lfns, reverse=True), 500):
        for storageElementName in storageElementNames:
            res = dm.removeReplica(storageElementName, lfnList)
            if not res["OK"]:
                print("Error:", res["Message"])
                continue
            for lfn in sorted(res["Value"]["Successful"]):
                print("Successfully removed %s replica of %s" % (storageElementName, lfn))
            for lfn in sorted(res["Value"]["Failed"]):
                message = res["Value"]["Failed"][lfn]
                print("Error: failed to remove %s replica of %s: %s" % (storageElementName, lfn, message))


if __name__ == "__main__":
    main()



""" NOTA BENE: This agent should NOT be run alone. Instead, it serves as a base class for extensions.

    The TaskManagerAgentBase is the base class to submit tasks to external systems,
    monitor and update the tasks and file status in the transformation DB.

    This agent is extended in WorkflowTaskAgent and RequestTaskAgent.
    In case you want to further extend it you are required to follow the note on the
    initialize method and on the _getClients method.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time
import datetime
import concurrent.futures

from DIRAC import S_OK

from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Utilities.Dictionaries import breakDictionaryIntoChunks
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getDNForUsername, getUsernameForDN
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.TransformationSystem.Client.FileReport import FileReport
from DIRAC.TransformationSystem.Client.WorkflowTasks import WorkflowTasks
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.TransformationSystem.Agent.TransformationAgentsUtilities import TransformationAgentsUtilities
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.Client.JobManagerClient import JobManagerClient

AGENT_NAME = "Transformation/TaskManagerAgentBase"


class TaskManagerAgentBase(AgentModule, TransformationAgentsUtilities):
    """To be extended. Please look at WorkflowTaskAgent and RequestTaskAgent."""

    def __init__(self, *args, **kwargs):
        """c'tor

        Always call this in the extension agent
        """
        AgentModule.__init__(self, *args, **kwargs)
        TransformationAgentsUtilities.__init__(self)

        self.transClient = None
        self.jobManagerClient = None
        self.transType = []

        self.tasksPerLoop = 50
        self.maxParametricJobs = 20  # will be updated in execute()

        # credentials
        self.shifterProxy = None
        self.credentials = None
        self.credTuple = (None, None, None)

        self.pluginLocation = ""
        self.bulkSubmissionFlag = False

    #############################################################################

    def initialize(self):
        """Agent initialization.

        The extensions MUST provide in the initialize method the following data members:
        - TransformationClient objects (self.transClient),
        - set the shifterProxy if different from the default one set here ('ProductionManager')
        - list of transformation types to be looked (self.transType)
        """

        gMonitor.registerActivity(
            "SubmittedTasks", "Automatically submitted tasks", "Transformation Monitoring", "Tasks", gMonitor.OP_ACUM
        )

        self.pluginLocation = self.am_getOption("PluginLocation", "DIRAC.TransformationSystem.Client.TaskManagerPlugin")

        # Default clients
        self.transClient = TransformationClient()
        self.jobManagerClient = JobManagerClient()

        # Bulk submission flag
        self.bulkSubmissionFlag = self.am_getOption("BulkSubmission", self.bulkSubmissionFlag)

        # Shifter credentials to use, could replace the use of shifterProxy eventually
        self.shifterProxy = self.am_getOption("shifterProxy", self.shifterProxy)
        self.credentials = self.am_getOption("ShifterCredentials", self.credentials)
        resCred = self.__getCredentials()
        if not resCred["OK"]:
            return resCred
        # setting up the threading
        maxNumberOfThreads = self.am_getOption("maxNumberOfThreads", 15)
        self.log.verbose("Multithreaded with %d threads" % maxNumberOfThreads)

        self.threadPoolExecutor = concurrent.futures.ThreadPoolExecutor(max_workers=maxNumberOfThreads)

        return S_OK()

    def finalize(self):
        """graceful finalization"""
        method = "finalize"
        self._logInfo("Wait for threads to get empty before terminating the agent", method=method)
        self.threadPoolExecutor.shutdown()
        self._logInfo("Threads are empty, terminating the agent...", method=method)
        return S_OK()

    def execute(self):
        """The execution method is transformations that need to be processed"""

        # 1. determining which credentials will be used for the submission
        owner, ownerGroup, ownerDN = None, None, None
        # getting the credentials for submission
        resProxy = getProxyInfo(proxy=False, disableVOMS=False)
        if resProxy["OK"]:  # there is a shifterProxy
            proxyInfo = resProxy["Value"]
            owner = proxyInfo["username"]
            ownerGroup = proxyInfo["group"]
            ownerDN = proxyInfo["identity"]
            self.log.info("ShifterProxy: Tasks will be submitted with the credentials %s:%s" % (owner, ownerGroup))
        elif self.credentials:
            owner, ownerGroup, ownerDN = self.credTuple
        else:
            self.log.info("Using per Transformation Credentials!")

        # 2. Determining which operations to do on each transformation
        self.operationsOnTransformationDict = {}  # key: TransID. Value: dict with body, and list of operations

        # 2.1 Determine whether the task status is to be monitored and updated
        if not self.am_getOption("MonitorTasks", ""):
            self.log.verbose("Monitoring of tasks is disabled. To enable it, create the 'MonitorTasks' option")
        else:
            # Get the transformations for which the tasks have to be updated
            status = self.am_getOption(
                "UpdateTasksTransformationStatus",
                self.am_getOption("UpdateTasksStatus", ["Active", "Completing", "Stopped"]),
            )
            transformations = self._selectTransformations(transType=self.transType, status=status, agentType=[])
            if not transformations["OK"]:
                self.log.warn("Could not select transformations:", transformations["Message"])
            else:
                self._addOperationForTransformations(
                    self.operationsOnTransformationDict,
                    "updateTaskStatus",
                    transformations,
                    owner=owner,
                    ownerGroup=ownerGroup,
                    ownerDN=ownerDN,
                )

        # 2.2. Determine whether the task files status is to be monitored and updated
        if not self.am_getOption("MonitorFiles", ""):
            self.log.verbose("Monitoring of files is disabled. To enable it, create the 'MonitorFiles' option")
        else:
            # Get the transformations for which the files have to be updated
            status = self.am_getOption(
                "UpdateFilesTransformationStatus",
                self.am_getOption("UpdateFilesStatus", ["Active", "Completing", "Stopped"]),
            )
            transformations = self._selectTransformations(transType=self.transType, status=status, agentType=[])
            if not transformations["OK"]:
                self.log.warn("Could not select transformations:", transformations["Message"])
            else:
                self._addOperationForTransformations(
                    self.operationsOnTransformationDict,
                    "updateFileStatus",
                    transformations,
                    owner=owner,
                    ownerGroup=ownerGroup,
                    ownerDN=ownerDN,
                )

        # Determine whether the checking of reserved tasks is to be performed
        if not self.am_getOption("CheckReserved", ""):
            self.log.verbose("Checking of reserved tasks is disabled. To enable it, create the 'CheckReserved' option")
        else:
            # Get the transformations for which the check of reserved tasks have to be performed
            status = self.am_getOption(
                "CheckReservedTransformationStatus",
                self.am_getOption("CheckReservedStatus", ["Active", "Completing", "Stopped"]),
            )
            transformations = self._selectTransformations(transType=self.transType, status=status, agentType=[])
            if not transformations["OK"]:
                self.log.warn("Could not select transformations:", transformations["Message"])
            else:
                self._addOperationForTransformations(
                    self.operationsOnTransformationDict,
                    "checkReservedTasks",
                    transformations,
                    owner=owner,
                    ownerGroup=ownerGroup,
                    ownerDN=ownerDN,
                )

        # Determine whether the submission of tasks is to be performed
        if not self.am_getOption("SubmitTasks", "yes"):
            self.log.verbose("Submission of tasks is disabled. To enable it, create the 'SubmitTasks' option")
        else:
            # Get the transformations for which the submission of tasks have to be performed
            status = self.am_getOption(
                "SubmitTransformationStatus", self.am_getOption("SubmitStatus", ["Active", "Completing"])
            )
            transformations = self._selectTransformations(transType=self.transType, status=status)
            if not transformations["OK"]:
                self.log.warn("Could not select transformations:", transformations["Message"])
            else:
                # Get the transformations which should be submitted
                self.tasksPerLoop = self.am_getOption("TasksPerLoop", self.tasksPerLoop)
                res = self.jobManagerClient.getMaxParametricJobs()
                if not res["OK"]:
                    self.log.warn("Could not get the maxParametricJobs from JobManager", res["Message"])
                else:
                    self.maxParametricJobs = res["Value"]

                self._addOperationForTransformations(
                    self.operationsOnTransformationDict,
                    "submitTasks",
                    transformations,
                    owner=owner,
                    ownerGroup=ownerGroup,
                    ownerDN=ownerDN,
                )

        # now call _execute...
        future_to_transID = {}
        for transID, transDict in self.operationsOnTransformationDict.items():
            future = self.threadPoolExecutor.submit(self._execute, transDict)
            future_to_transID[future] = transID

        for future in concurrent.futures.as_completed(future_to_transID):
            transID = future_to_transID[future]
            try:
                future.result()
            except Exception as exc:
                self._logError("%d generated an exception: %s" % (transID, exc))
            else:
                self._logInfo("Processed %d" % transID)

        return S_OK()

    def _selectTransformations(self, transType=None, status=None, agentType=None):
        """get the transformations"""
        if status is None:
            status = ["Active", "Completing"]
        if agentType is None:
            agentType = ["Automatic"]
        selectCond = {}
        if status:
            selectCond["Status"] = status
        if transType is not None:
            selectCond["Type"] = transType
        if agentType:
            selectCond["AgentType"] = agentType
        res = self.transClient.getTransformations(condDict=selectCond)
        if not res["OK"]:
            self.log.error("Failed to get transformations:", res["Message"])
        elif not res["Value"]:
            self.log.verbose("No transformations found")
        else:
            self.log.verbose("Obtained %d transformations" % len(res["Value"]))
        return res

    #############################################################################

    def _getClients(self, ownerDN=None, ownerGroup=None):
        """Returns the clients used in the threads

        This is another function that should be extended.

        The clients provided here are defaults, and should be adapted

        If ownerDN and ownerGroup are not None the clients will delegate to these credentials

        :param str ownerDN: DN of the owner of the submitted jobs
        :param str ownerGroup: group of the owner of the submitted jobs
        :returns: dict of Clients
        """
        threadTransformationClient = TransformationClient()
        threadTaskManager = WorkflowTasks(ownerDN=ownerDN, ownerGroup=ownerGroup)
        threadTaskManager.pluginLocation = self.pluginLocation

        return {"TransformationClient": threadTransformationClient, "TaskManager": threadTaskManager}

    def _execute(self, transDict):
        """This is what runs inside the threads, in practice this is the function that does the real stuff"""
        # Each thread will have its own clients if we use credentials/shifterProxy
        clients = (
            self._getClients()
            if self.shifterProxy
            else self._getClients(ownerGroup=self.credTuple[1], ownerDN=self.credTuple[2])
            if self.credentials
            else None
        )

        method = "_execute"
        operation = "None"

        startTime = time.time()

        try:
            transID = transDict["TransformationID"]
            operations = transDict["Operations"]
            if not (self.credentials or self.shifterProxy):
                ownerDN, group = transDict["OwnerDN"], transDict["OwnerGroup"]
                clients = self._getClients(ownerDN=ownerDN, ownerGroup=group)
            self._logInfo("Start processing transformation", method=method, transID=transID)
            for operation in operations:
                self._logInfo("Executing %s" % operation, method=method, transID=transID)
                startOperation = time.time()
                res = getattr(self, operation)(transDict, clients)
                if not res["OK"]:
                    self._logError(
                        "Failed to execute '%s': %s" % (operation, res["Message"]), method=method, transID=transID
                    )
                self._logInfo(
                    "Executed %s in %.1f seconds" % (operation, time.time() - startOperation),
                    method=method,
                    transID=transID,
                )
        except Exception as x:  # pylint: disable=broad-except
            self._logException(
                "Exception executing operation %s" % operation, lException=x, method=method, transID=transID
            )
        finally:
            self._logInfo(
                "Processed transformation in %.1f seconds" % (time.time() - startTime), method=method, transID=transID
            )

    #############################################################################
    # real operations done

    def updateTaskStatus(self, transDict, clients):
        """Updates the task status"""
        transID = transDict["TransformationID"]
        method = "updateTaskStatus"

        # Get the tasks which are in an UPDATE state, i.e. job statuses + request-specific statuses
        updateStatus = self.am_getOption(
            "TaskUpdateStatus",
            [
                JobStatus.CHECKING,
                JobStatus.DELETED,
                JobStatus.KILLED,
                JobStatus.STAGING,
                JobStatus.STALLED,
                JobStatus.MATCHED,
                JobStatus.RESCHEDULED,
                JobStatus.COMPLETING,
                JobStatus.COMPLETED,
                JobStatus.SUBMITTING,
                JobStatus.RECEIVED,
                JobStatus.WAITING,
                JobStatus.RUNNING,
                "Scheduled",
                "Assigned",
            ],
        )
        condDict = {"TransformationID": transID, "ExternalStatus": updateStatus}
        timeStamp = str(datetime.datetime.utcnow() - datetime.timedelta(minutes=10))

        # Get transformation tasks
        transformationTasks = clients["TransformationClient"].getTransformationTasks(
            condDict=condDict, older=timeStamp, timeStamp="LastUpdateTime"
        )
        if not transformationTasks["OK"]:
            self._logError(
                "Failed to get tasks to update:", transformationTasks["Message"], method=method, transID=transID
            )
            return transformationTasks
        if not transformationTasks["Value"]:
            self._logVerbose("No tasks found to update", method=method, transID=transID)
            return transformationTasks

        # Get status for the transformation tasks
        chunkSize = self.am_getOption("TaskUpdateChunkSize", 0)
        try:
            chunkSize = int(chunkSize)
        except ValueError:
            chunkSize = 0
        if chunkSize:
            self._logVerbose(
                "Getting %d tasks status (chunks of %d)" % (len(transformationTasks["Value"]), chunkSize),
                method=method,
                transID=transID,
            )
        else:
            self._logVerbose(
                "Getting %d tasks status" % len(transformationTasks["Value"]), method=method, transID=transID
            )
        updated = {}
        for nb, taskChunk in enumerate(
            breakListIntoChunks(transformationTasks["Value"], chunkSize)
            if chunkSize
            else [transformationTasks["Value"]]
        ):
            submittedTaskStatus = clients["TaskManager"].getSubmittedTaskStatus(taskChunk)
            if not submittedTaskStatus["OK"]:
                self._logError(
                    "Failed to get updated task states:", submittedTaskStatus["Message"], method=method, transID=transID
                )
                return submittedTaskStatus
            statusDict = submittedTaskStatus["Value"]
            if not statusDict:
                self._logVerbose("%4d: No tasks to update" % nb, method=method, transID=transID)

            # Set status for tasks that changes
            for status, taskIDs in statusDict.items():
                self._logVerbose(
                    "%4d: Updating %d task(s) to %s" % (nb, len(taskIDs), status), method=method, transID=transID
                )
                setTaskStatus = clients["TransformationClient"].setTaskStatus(transID, taskIDs, status)
                if not setTaskStatus["OK"]:
                    self._logError(
                        "Failed to update task status for transformation:",
                        setTaskStatus["Message"],
                        method=method,
                        transID=transID,
                    )
                    return setTaskStatus
                updated[status] = updated.setdefault(status, 0) + len(taskIDs)

        for status, nb in updated.items():
            self._logInfo("Updated %d tasks to status %s" % (nb, status), method=method, transID=transID)
        return S_OK()

    def updateFileStatus(self, transDict, clients):
        """Update the files status"""
        transID = transDict["TransformationID"]
        method = "updateFileStatus"

        timeStamp = str(datetime.datetime.utcnow() - datetime.timedelta(minutes=10))

        # get transformation files
        condDict = {"TransformationID": transID, "Status": ["Assigned"]}
        transformationFiles = clients["TransformationClient"].getTransformationFiles(
            condDict=condDict, older=timeStamp, timeStamp="LastUpdate"
        )
        if not transformationFiles["OK"]:
            self._logError(
                "Failed to get transformation files to update:",
                transformationFiles["Message"],
                method=method,
                transID=transID,
            )
            return transformationFiles
        if not transformationFiles["Value"]:
            self._logInfo("No files to be updated", method=method, transID=transID)
            return transformationFiles

        # Get the status of the transformation files
        # Sort the files by taskID
        taskFiles = {}
        for fileDict in transformationFiles["Value"]:
            taskFiles.setdefault(fileDict["TaskID"], []).append(fileDict)

        chunkSize = 100
        self._logVerbose(
            "Getting file status for %d tasks (chunks of %d)" % (len(taskFiles), chunkSize),
            method=method,
            transID=transID,
        )
        updated = {}
        # Process 100 tasks at a time
        for nb, taskIDs in enumerate(breakListIntoChunks(taskFiles, chunkSize)):
            fileChunk = []
            for taskID in taskIDs:
                fileChunk += taskFiles[taskID]
            submittedFileStatus = clients["TaskManager"].getSubmittedFileStatus(fileChunk)
            if not submittedFileStatus["OK"]:
                self._logError(
                    "Failed to get updated file states for transformation:",
                    submittedFileStatus["Message"],
                    method=method,
                    transID=transID,
                )
                return submittedFileStatus
            statusDict = submittedFileStatus["Value"]
            if not statusDict:
                self._logVerbose("%4d: No file states to be updated" % nb, method=method, transID=transID)
                continue

            # Set the status of files
            fileReport = FileReport(server=clients["TransformationClient"].getServer())
            for lfn, status in statusDict.items():
                updated[status] = updated.setdefault(status, 0) + 1
                setFileStatus = fileReport.setFileStatus(transID, lfn, status)
                if not setFileStatus["OK"]:
                    return setFileStatus
            commit = fileReport.commit()
            if not commit["OK"]:
                self._logError(
                    "Failed to update file states for transformation:",
                    commit["Message"],
                    method=method,
                    transID=transID,
                )
                return commit
            else:
                self._logVerbose(
                    "%4d: Updated the states of %d files" % (nb, len(commit["Value"])), method=method, transID=transID
                )

        for status, nb in updated.items():
            self._logInfo("Updated %d files to status %s" % (nb, status), method=method, transID=transID)
        return S_OK()

    def checkReservedTasks(self, transDict, clients):
        """Checking Reserved tasks"""
        transID = transDict["TransformationID"]
        method = "checkReservedTasks"

        # Select the tasks which have been in Reserved status for more than 1 hour for selected transformations
        condDict = {"TransformationID": transID, "ExternalStatus": "Reserved"}
        time_stamp_older = str(datetime.datetime.utcnow() - datetime.timedelta(hours=1))

        res = clients["TransformationClient"].getTransformationTasks(condDict=condDict, older=time_stamp_older)
        self._logDebug("getTransformationTasks(%s) return value:" % condDict, res, method=method, transID=transID)
        if not res["OK"]:
            self._logError("Failed to get Reserved tasks:", res["Message"], method=method, transID=transID)
            return res
        if not res["Value"]:
            self._logVerbose("No Reserved tasks found", transID=transID)
            return res
        reservedTasks = res["Value"]

        # Update the reserved tasks
        res = clients["TaskManager"].updateTransformationReservedTasks(reservedTasks)
        self._logDebug(
            "updateTransformationReservedTasks(%s) return value:" % reservedTasks, res, method=method, transID=transID
        )
        if not res["OK"]:
            self._logError(
                "Failed to update transformation reserved tasks:", res["Message"], method=method, transID=transID
            )
            return res
        noTasks = res["Value"]["NoTasks"]
        taskNameIDs = res["Value"]["TaskNameIDs"]

        # For the tasks with no associated request found re-set the status of the task in the transformationDB
        if noTasks:
            self._logInfo(
                "Resetting status of %d tasks to Created as no associated job/request found" % len(noTasks),
                method=method,
                transID=transID,
            )
            for taskName in noTasks:
                transID, taskID = self._parseTaskName(taskName)
                res = clients["TransformationClient"].setTaskStatus(transID, taskID, "Created")
                if not res["OK"]:
                    self._logError(
                        "Failed to update task status and ID after recovery:",
                        "%s %s" % (taskName, res["Message"]),
                        method=method,
                        transID=transID,
                    )
                    return res

        # For the tasks for which an associated request was found update the task details in the transformationDB
        for taskName, extTaskID in taskNameIDs.items():
            transID, taskID = self._parseTaskName(taskName)
            self._logInfo(
                "Setting status of %s to Submitted with ID %s" % (taskName, extTaskID), method=method, transID=transID
            )
            setTaskStatusAndWmsID = clients["TransformationClient"].setTaskStatusAndWmsID(
                transID, taskID, "Submitted", str(extTaskID)
            )
            if not setTaskStatusAndWmsID["OK"]:
                self._logError(
                    "Failed to update task status and ID after recovery:",
                    "%s %s" % (taskName, setTaskStatusAndWmsID["Message"]),
                    method=method,
                    transID=transID,
                )
                return setTaskStatusAndWmsID

        return S_OK()

    def submitTasks(self, transDict, clients):
        """Submit the tasks to an external system, using the taskManager provided

        :param dict transIDOPBody: transformation body
        :param dict clients: dictionary of client objects

        :return: S_OK/S_ERROR
        """
        transID = transDict["TransformationID"]
        transBody = transDict["Body"]
        owner = transDict["Owner"]
        ownerGroup = transDict["OwnerGroup"]
        ownerDN = transDict["OwnerDN"]
        method = "submitTasks"

        # Get all tasks to submit
        tasksToSubmit = clients["TransformationClient"].getTasksToSubmit(transID, self.tasksPerLoop)
        self._logDebug(
            "getTasksToSubmit(%s, %s) return value:" % (transID, self.tasksPerLoop),
            tasksToSubmit,
            method=method,
            transID=transID,
        )
        if not tasksToSubmit["OK"]:
            self._logError("Failed to obtain tasks:", tasksToSubmit["Message"], method=method, transID=transID)
            return tasksToSubmit
        tasks = tasksToSubmit["Value"]["JobDictionary"]
        if not tasks:
            self._logVerbose("No tasks found for submission", method=method, transID=transID)
            return tasksToSubmit
        self._logInfo("Obtained %d tasks for submission" % len(tasks), method=method, transID=transID)

        # Prepare tasks and submits them, by chunks
        chunkSize = self.maxParametricJobs if self.bulkSubmissionFlag else self.tasksPerLoop
        for taskDictChunk in breakDictionaryIntoChunks(tasks, chunkSize):
            res = self._prepareAndSubmitAndUpdateTasks(
                transID, transBody, taskDictChunk, owner, ownerDN, ownerGroup, clients
            )
            if not res["OK"]:
                return res
            self._logVerbose(
                "Submitted %d jobs, bulkSubmissionFlag = %s" % (len(taskDictChunk), self.bulkSubmissionFlag)
            )

        return S_OK()

    def _prepareAndSubmitAndUpdateTasks(self, transID, transBody, tasks, owner, ownerDN, ownerGroup, clients):
        """prepare + submit + monitor a dictionary of tasks

        :param int transID: transformation ID
        :param str transBody: transformation job template
        :param dict tasks: dictionary of per task parameters
        :param str owner: owner of the transformation
        :param str ownerDN: DN of the owner of the transformation
        :param str ownerGroup: group of the owner of the transformation
        :param dict clients: dictionary of client objects

        :return: S_OK/S_ERROR
        """

        method = "_prepareAndSubmitAndUpdateTasks"
        # prepare tasks
        preparedTransformationTasks = clients["TaskManager"].prepareTransformationTasks(
            transBody, tasks, owner, ownerGroup, ownerDN, self.bulkSubmissionFlag
        )
        self._logDebug(
            "prepareTransformationTasks return value:", preparedTransformationTasks, method=method, transID=transID
        )
        if not preparedTransformationTasks["OK"]:
            self._logError(
                "Failed to prepare tasks", preparedTransformationTasks["Message"], method=method, transID=transID
            )
            return preparedTransformationTasks

        # Submit tasks
        res = clients["TaskManager"].submitTransformationTasks(preparedTransformationTasks["Value"])
        self._logDebug("submitTransformationTasks return value:", res, method=method, transID=transID)
        if not res["OK"]:
            self._logError("Failed to submit prepared tasks:", res["Message"], method=method, transID=transID)
            return res

        # Update tasks after submission
        res = clients["TaskManager"].updateDBAfterTaskSubmission(res["Value"])
        self._logDebug("updateDBAfterTaskSubmission return value:", res, method=method, transID=transID)
        if not res["OK"]:
            self._logError("Failed to update DB after task submission:", res["Message"], method=method, transID=transID)
            return res

        return S_OK()

    @staticmethod
    def _addOperationForTransformations(
        operationsOnTransformationDict,
        operation,
        transformations,
        owner=None,
        ownerGroup=None,
        ownerDN=None,
    ):
        """Fill the operationsOnTransformationDict"""

        transformationIDsAndBodies = (
            (
                transformation["TransformationID"],
                transformation["Body"],
                transformation["AuthorDN"],
                transformation["AuthorGroup"],
            )
            for transformation in transformations["Value"]
        )
        for transID, body, t_ownerDN, t_ownerGroup in transformationIDsAndBodies:
            if transID in operationsOnTransformationDict:
                operationsOnTransformationDict[transID]["Operations"].append(operation)
            else:
                operationsOnTransformationDict[transID] = {
                    "TransformationID": transID,
                    "Body": body,
                    "Operations": [operation],
                    "Owner": owner if owner else getUsernameForDN(t_ownerDN)["Value"],
                    "OwnerGroup": ownerGroup if owner else t_ownerGroup,
                    "OwnerDN": ownerDN if owner else t_ownerDN,
                }

    def __getCredentials(self):
        """Get the credentials to use if ShifterCredentials are set, otherwise do nothing.

        This function fills the self.credTuple tuple.
        """
        if not self.credentials:
            return S_OK()
        resCred = Operations().getOptionsDict("/Shifter/%s" % self.credentials)
        if not resCred["OK"]:
            self.log.error("Cred: Failed to find shifter credentials", self.credentials)
            return resCred
        owner = resCred["Value"]["User"]
        ownerGroup = resCred["Value"]["Group"]
        # returns  a list
        ownerDN = getDNForUsername(owner)["Value"][0]
        self.credTuple = (owner, ownerGroup, ownerDN)
        self.log.info("Cred: Tasks will be submitted with the credentials %s:%s" % (owner, ownerGroup))
        return S_OK()



""" Just listing the possible Properties
This module contains list of Properties that can be assigned to users and groups
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


#: A host property. This property is used::
#: * For a host to forward credentials in a DISET call
TRUSTED_HOST = "TrustedHost"

#: Normal user operations
NORMAL_USER = "NormalUser"

#: CS Administrator - possibility to edit the Configuration Service
CS_ADMINISTRATOR = "CSAdministrator"

#: Job sharing among members of a group
JOB_SHARING = "JobSharing"

#: DIRAC Service Administrator
SERVICE_ADMINISTRATOR = "ServiceAdministrator"

#: Job Administrator can manipulate everybody's jobs
JOB_ADMINISTRATOR = "JobAdministrator"

#: Job Monitor - can get job monitoring information
JOB_MONITOR = "JobMonitor"

#: Accounting Monitor - can see accounting data for all groups
ACCOUNTING_MONITOR = "AccountingMonitor"

#: Private pilot
PILOT = "Pilot"

#: Generic pilot
GENERIC_PILOT = "GenericPilot"

#: Site Manager
SITE_MANAGER = "SiteManager"

#: User, group, VO Registry management
USER_MANAGER = "UserManager"

#: Operator
OPERATOR = "Operator"

#: Allow getting full delegated proxies
FULL_DELEGATION = "FullDelegation"

#: Allow getting only limited proxies (ie. pilots)
LIMITED_DELEGATION = "LimitedDelegation"

#: Allow getting only limited proxies for one self
PRIVATE_LIMITED_DELEGATION = "PrivateLimitedDelegation"

#: Allow managing proxies
PROXY_MANAGEMENT = "ProxyManagement"

#: Allow managing production
PRODUCTION_MANAGEMENT = "ProductionManagement"

#: Allow production request approval on behalf of PPG
PPG_AUTHORITY = "PPGAuthority"

#: Allow Bookkeeping Management
BOOKKEEPING_MANAGEMENT = "BookkeepingManagement"

#: Allow to set notifications and manage alarms
ALARMS_MANAGEMENT = "AlarmsManagement"

#: Allow FC Management - FC root user
FC_MANAGEMENT = "FileCatalogManagement"

#: Allow staging files
STAGE_ALLOWED = "StageAllowed"

#: Allow VMDIRAC Operations via various handlers
VM_RPC_OPERATION = "VmRpcOperation"



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.AccountingSystem.Client.Types.DataOperation import DataOperation
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter


class DataOperationPlotter(BaseReporter):

    _typeName = "DataOperation"
    _typeKeyFields = [dF[0] for dF in DataOperation().definitionKeyFields]

    def _translateGrouping(self, grouping):
        if grouping == "Channel":
            return ("%s, %s", ["Source", "Destination"], "CONCAT( %s, ' -> ', %s )")
        else:
            return ("%s", [grouping])

    _reportSuceededTransfersName = "Successful transfers"

    def _reportSuceededTransfers(self, reportRequest):
        return self.__reportTransfers(reportRequest, "Succeeded", ("Failed", 0))

    _reportFailedTransfersName = "Failed transfers"

    def _reportFailedTransfers(self, reportRequest):
        return self.__reportTransfers(reportRequest, "Failed", ("Succeeded", 1))

    def __reportTransfers(self, reportRequest, titleType, togetherFieldsToPlot):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s), SUM(%s)-SUM(%s)",
            reportRequest["groupingFields"][1]
            + [
                "startTime",
                "bucketLength",
                "TransferOK",
                "TransferTotal",
                "TransferOK",
            ],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        strippedData = self.stripDataField(dataDict, togetherFieldsToPlot[1])
        if strippedData:
            dataDict[togetherFieldsToPlot[0]] = strippedData[0]
        dataDict, maxValue = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, maxValue, unitName = self._findSuitableRateUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "files"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotSuceededTransfers(self, reportRequest, plotInfo, filename):
        return self.__plotTransfers(reportRequest, plotInfo, filename, "Succeeded", ("Failed", 0))

    def _plotFailedTransfers(self, reportRequest, plotInfo, filename):
        return self.__plotTransfers(reportRequest, plotInfo, filename, "Failed", ("Succeeded", 1))

    def __plotTransfers(self, reportRequest, plotInfo, filename, titleType, togetherFieldsToPlot):
        metadata = {
            "title": "%s Transfers by %s" % (titleType, reportRequest["grouping"]),
            "ylabel": plotInfo["unit"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
        }
        return self._generateTimedStackedBarPlot(filename, plotInfo["graphDataDict"], metadata)

    _reportQualityName = "Efficiency by protocol"

    def _reportQuality(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s), SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "TransferOK", "TransferTotal"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {
                "checkNone": True,
                "convertToGranularity": "sum",
                "calculateProportionalGauges": False,
                "consolidationFunction": self._efficiencyConsolidation,
            },
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        if len(dataDict) > 1:
            # Get the total for the plot
            selectFields = (
                "'Total', %s, %s, SUM(%s),SUM(%s)",
                ["startTime", "bucketLength", "TransferOK", "TransferTotal"],
            )
            retVal = self._getTimedData(
                reportRequest["startTime"],
                reportRequest["endTime"],
                selectFields,
                reportRequest["condDict"],
                reportRequest["groupingFields"],
                {
                    "checkNone": True,
                    "convertToGranularity": "sum",
                    "calculateProportionalGauges": False,
                    "consolidationFunction": self._efficiencyConsolidation,
                },
            )
            if not retVal["OK"]:
                return retVal
            totalDict = retVal["Value"][0]
            self.stripDataField(totalDict, 0)
            for key in totalDict:
                dataDict[key] = totalDict[key]
        return S_OK({"data": dataDict, "granularity": granularity})

    def _plotQuality(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Transfer quality by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
        }
        return self._generateQualityPlot(filename, plotInfo["data"], metadata)

    _reportTransferedDataName = "Cumulative transferred data"

    def _reportTransferedData(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "TransferSize"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        dataDict = self._accumulate(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, maxValue, unitName = self._findSuitableUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "bytes"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotTransferedData(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Transfered data by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
            "sort_labels": "last_value",
        }
        return self._generateCumulativePlot(filename, plotInfo["graphDataDict"], metadata)

    def _reportThroughput(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "TransferSize"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict, maxValue = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, maxValue, unitName = self._findSuitableRateUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "bytes"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotThroughput(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Throughput by %s" % reportRequest["grouping"],
            "ylabel": plotInfo["unit"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
        }
        return self._generateTimedStackedBarPlot(filename, plotInfo["graphDataDict"], metadata)

    _reportDataTransferedName = "Pie chart of transferred data"

    def _reportDataTransfered(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", SUM(%s)",
            reportRequest["groupingFields"][1] + ["TransferSize"],
        )
        retVal = self._getSummaryData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict = retVal["Value"]
        for key in dataDict:
            dataDict[key] = int(dataDict[key])
        return S_OK({"data": dataDict})

    def _plotDataTransfered(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Total data transfered by %s" % reportRequest["grouping"],
            "ylabel": "bytes",
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
        }
        return self._generatePiePlot(filename, plotInfo["data"], metadata)



""" DIRAC FileCatalog Security Manager mix-in class for access check only on the directory level
    with a special treatment of the Delete operation
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.SecurityManager.DirectorySecurityManager import (
    DirectorySecurityManager,
)
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.SecurityManager.SecurityManagerBase import (
    _readMethods,
    _writeMethods,
)


class DirectorySecurityManagerWithDelete(DirectorySecurityManager):
    """This security manager implements a Delete operation.
    For Read, Write, Execute, it's behavior is the one of DirectorySecurityManager.
    For Delete, if the directory does not exist, we return True.
    If the directory exists, then we test the Write permission

    """

    def hasAccess(self, opType, paths, credDict):
        # The other SecurityManager do not support the Delete operation,
        # and it is transformed in Write
        # so we keep the original one

        if opType in ["removeFile", "removeReplica", "removeDirectory"]:
            self.opType = "Delete"
        elif opType in _readMethods:
            self.opType = "Read"
        elif opType in _writeMethods:
            self.opType = "Write"

        res = super(DirectorySecurityManagerWithDelete, self).hasAccess(opType, paths, credDict)

        # We reinitialize self.opType in case someone would call getPathPermissions directly
        self.opType = ""

        return res

    def getPathPermissions(self, paths, credDict):
        """Get path permissions according to the policy"""

        # If we are testing in anything else than a Delete, just return the parent methods
        if hasattr(self, "opType") and self.opType.lower() != "delete":
            return super(DirectorySecurityManagerWithDelete, self).getPathPermissions(paths, credDict)

        # If the object (file or dir) does not exist, we grant the permission
        res = self.db.dtree.exists(paths)
        if not res["OK"]:
            return res

        nonExistingDirectories = set(
            path for path in res["Value"]["Successful"] if not res["Value"]["Successful"][path]
        )

        res = self.db.fileManager.exists(paths)
        if not res["OK"]:
            return res

        nonExistingFiles = set(path for path in res["Value"]["Successful"] if not res["Value"]["Successful"][path])

        nonExistingObjects = nonExistingDirectories & nonExistingFiles

        permissions = {}
        failed = {}

        for path in nonExistingObjects:
            permissions[path] = {"Read": True, "Write": True, "Execute": True}
            # The try catch is just to protect in case there are duplicate in the paths
            try:
                paths.remove(path)
            except Exception as _e:
                try:
                    paths.pop(path)
                except Exception as _ee:
                    pass

        # For all the paths that exist, check the write permission
        if paths:

            res = super(DirectorySecurityManagerWithDelete, self).getPathPermissions(paths, credDict)
            if not res["OK"]:
                return res

            failed = res["Value"]["Failed"]
            permissions.update(res["Value"]["Successful"])

        return S_OK({"Successful": permissions, "Failed": failed})



""" PlottingClient is a client of the Plotting Service
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import tempfile
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.Core.DISET.TransferClient import TransferClient


class PlottingClient(object):
    def __init__(self, rpcClient=None, transferClient=None):
        self.serviceName = "Framework/Plotting"
        self.rpcClient = rpcClient
        self.transferClient = transferClient

    def __getRPCClient(self):
        if self.rpcClient:
            return self.rpcClient
        return RPCClient(self.serviceName)

    def __getTransferClient(self):
        if self.transferClient:
            return self.transferClient
        return TransferClient(self.serviceName)

    def getPlotToMemory(self, plotName):
        """Get the prefabricated plot from the service and return it as a string"""
        transferClient = self.__getTransferClient()
        tmpFile = tempfile.TemporaryFile()
        retVal = transferClient.receiveFile(tmpFile, plotName)
        if not retVal["OK"]:
            return retVal
        tmpFile.seek(0)
        data = tmpFile.read()
        tmpFile.close()
        return S_OK(data)

    def getPlotToFile(self, plotName, fileName):
        """Get the prefabricated plot from the service and store it in a file"""
        transferClient = self.__getTransferClient()
        try:
            with open(fileName, "wb") as destFile:
                retVal = transferClient.receiveFile(destFile, plotName)
        except Exception as e:
            return S_ERROR("Can't open file %s for writing: %s" % (fileName, str(e)))
        if not retVal["OK"]:
            return retVal
        return S_OK(fileName)

    def graph(self, data, fname=False, *args, **kw):
        """Generic method to obtain graphs from the Plotting service. The requested
        graphs are completely described by their data and metadata
        """

        client = self.__getRPCClient()
        plotMetadata = {}
        for arg in args:
            if isinstance(arg, dict):
                plotMetadata.update(arg)
            else:
                return S_ERROR("Non-dictionary non-keyed argument")
        plotMetadata.update(kw)
        result = client.generatePlot(data, plotMetadata)
        if not result["OK"]:
            return result

        plotName = result["Value"]
        if fname and fname != "Memory":
            result = self.getPlotToFile(plotName, fname)
        else:
            result = self.getPlotToMemory(plotName)

        return result

    def barGraph(self, data, fileName, *args, **kw):
        return self.graph(data, fileName, plot_type="BarGraph", statistics_line=True, *args, **kw)

    def lineGraph(self, data, fileName, *args, **kw):
        return self.graph(data, fileName, plot_type="LineGraph", statistics_line=True, *args, **kw)

    def curveGraph(self, data, fileName, *args, **kw):
        return self.graph(data, fileName, plot_type="CurveGraph", statistics_line=True, *args, **kw)

    def cumulativeGraph(self, data, fileName, *args, **kw):
        return self.graph(data, fileName, plot_type="LineGraph", cumulate_data=True, *args, **kw)

    def pieGraph(self, data, fileName, *args, **kw):
        prefs = {"xticks": False, "yticks": False, "legend_position": "right"}
        return self.graph(data, fileName, prefs, plot_type="PieGraph", *args, **kw)

    def qualityGraph(self, data, fileName, *args, **kw):
        prefs = {"plot_axis_grid": False}
        return self.graph(data, fileName, prefs, plot_type="QualityMapGraph", *args, **kw)

    def textGraph(self, text, fileName, *args, **kw):
        prefs = {"text_image": text}
        return self.graph({}, fileName, prefs, *args, **kw)

    def histogram(self, data, fileName, bins, *args, **kw):
        try:
            from pylab import hist
        except Exception:
            return S_ERROR("No pylab module available")
        values, vbins, patches = hist(data, bins)
        histo = dict(zip(vbins, values))
        span = (max(data) - min(data)) / float(bins) * 0.98
        return self.graph(histo, fileName, plot_type="BarGraph", span=span, statistics_line=True, *args, **kw)



""" EC2Endpoint class is the implementation of the EC2 interface to
    a cloud endpoint
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import json
import boto3

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.File import makeGuid
from DIRAC.Resources.Cloud.Endpoint import Endpoint

__RCSID__ = "$Id$"


class EC2Endpoint(Endpoint):
    def __init__(self, parameters=None):
        super(EC2Endpoint, self).__init__(parameters=parameters)
        # logger
        self.log = gLogger.getSubLogger("EC2Endpoint")
        self.valid = False
        result = self.initialize()
        if result["OK"]:
            self.log.debug("EC2Endpoint created and validated")
            self.valid = True
        else:
            self.log.error(result["Message"])

    def initialize(self):

        availableParams = {
            "RegionName": "region_name",
            "AccessKey": "aws_access_key_id",
            "SecretKey": "aws_secret_access_key",
            "EndpointUrl": "endpoint_url",  # EndpointUrl is optional
        }

        connDict = {}
        for var in availableParams:
            if var in self.parameters:
                connDict[availableParams[var]] = self.parameters[var]

        try:
            self.__ec2 = boto3.resource("ec2", **connDict)
        except Exception as e:
            self.log.exception("Failed to connect to EC2")
            errorStatus = "Can't connect to EC2: " + str(e)
            return S_ERROR(errorStatus)

        result = self.__loadInstanceType()
        if not result["OK"]:
            return result

        result = self.__checkConnection()
        return result

    def __loadInstanceType(self):
        currentDir = os.path.dirname(os.path.abspath(__file__))
        instanceTypeFile = os.path.join(currentDir, "ec2_instance_type.json")
        try:
            with open(instanceTypeFile, "r") as f:
                self.__instanceTypeInfo = json.load(f)
        except Exception as e:
            self.log.exception("Failed to fetch EC2 instance details")
            errmsg = "Exception loading EC2 instance type info: %s" % e
            self.log.error(errmsg)
            return S_ERROR(errmsg)

        return S_OK()

    def __checkConnection(self):
        """
        Checks connection status by trying to list the images.

        :return: S_OK | S_ERROR
        """
        try:
            self.__ec2.images.filter(Owners=["self"])
        except Exception as e:
            self.log.exception("Failed to list EC2 images")
            return S_ERROR(e)

        return S_OK()

    def createInstances(self, vmsToSubmit):
        outputDict = {}

        for nvm in range(vmsToSubmit):
            instanceID = makeGuid()[:8]
            result = self.createInstance(instanceID)
            if result["OK"]:
                ec2Id, nodeDict = result["Value"]
                self.log.debug("Created VM instance %s/%s" % (ec2Id, instanceID))
                outputDict[ec2Id] = nodeDict
            else:
                self.log.error("Create EC2 instance error:", result["Message"])
                break

        return S_OK(outputDict)

    def createInstance(self, instanceID=""):
        if not instanceID:
            instanceID = makeGuid()[:8]

        self.parameters["VMUUID"] = instanceID
        self.parameters["VMType"] = self.parameters.get("CEType", "EC2")

        createNodeDict = {}

        # Image
        if "ImageID" in self.parameters and "ImageName" not in self.parameters:
            try:
                images = self.__ec2.images.filter(Filters=[{"Name": "name", "Values": [self.parameters["ImageName"]]}])
                imageId = None
                for image in images:
                    imageId = image.id
                    break
            except Exception as e:
                self.log.exception("Exception when get ID from image name %s:" % self.parameters["ImageName"])
                return S_ERROR("Failed to get image for Name %s" % self.parameters["ImageName"])
            if imageId is None:
                return S_ERROR("Image name %s not found" % self.parameters["ImageName"])
        elif "ImageID" in self.parameters:
            try:
                self.__ec2.images.filter(ImageIds=[self.parameters["ImageID"]])
            except Exception as e:
                self.log.exception("Failed to get EC2 image list")
                return S_ERROR("Failed to get image for ID %s" % self.parameters["ImageID"])
            imageId = self.parameters["ImageID"]
        else:
            return S_ERROR("No image specified")
        createNodeDict["ImageId"] = imageId

        # Instance type
        if "FlavorName" not in self.parameters:
            return S_ERROR("No flavor specified")
        instanceType = self.parameters["FlavorName"]
        createNodeDict["InstanceType"] = instanceType

        # User data
        result = self._createUserDataScript()
        if not result["OK"]:
            return result
        createNodeDict["UserData"] = str(result["Value"])

        # Other params
        for param in ["KeyName", "SubnetId", "EbsOptimized"]:
            if param in self.parameters:
                createNodeDict[param] = self.parameters[param]

        self.log.info("Creating node:")
        for key, value in createNodeDict.items():
            self.log.verbose("%s: %s" % (key, value))

        # Create the VM instance now
        try:
            instances = self.__ec2.create_instances(MinCount=1, MaxCount=1, **createNodeDict)
        except Exception as e:
            self.log.exception("Failed to create EC2 instance")
            return S_ERROR("Exception in ec2 create_instances: %s" % e)

        if len(instances) < 1:
            errmsg = "ec2 create_instances failed to create any VM"
            self.log.error(errmsg)
            return S_ERROR(errmsg)

        # Create the name in tags
        ec2Id = instances[0].id
        tags = [{"Key": "Name", "Value": "DIRAC_%s" % instanceID}]
        try:
            self.__ec2.create_tags(Resources=[ec2Id], Tags=tags)
        except Exception as e:
            self.log.exception("Failed to tag EC2 instance")
            return S_ERROR("Exception setup name for %s: %s" % (ec2Id, e))

        # Properties of the instance
        nodeDict = {}
        #    nodeDict['PublicIP'] = publicIP
        nodeDict["InstanceID"] = instanceID
        if instanceType in self.__instanceTypeInfo:
            nodeDict["NumberOfProcessors"] = self.__instanceTypeInfo[instanceType]["vCPU"]
            nodeDict["RAM"] = self.__instanceTypeInfo[instanceType]["Memory"]
        else:
            nodeDict["NumberOfProcessors"] = 1

        return S_OK((ec2Id, nodeDict))

    def stopVM(self, nodeID, publicIP=""):
        """
        Given the node ID it gets the node details, which are used to destroy the
        node making use of the libcloud.openstack driver. If three is any public IP
        ( floating IP ) assigned, frees it as well.

        :Parameters:
          **uniqueId** - `string`
            openstack node id ( not uuid ! )
          **public_ip** - `string`
            public IP assigned to the node if any

        :return: S_OK | S_ERROR
        """
        try:
            self.__ec2.Instance(nodeID).terminate()
        except Exception as e:
            self.log.exception("Failed to terminate EC2 instance")
            return S_ERROR("Exception terminate instance %s: %s" % (nodeID, e))

        return S_OK()



""" DowntimePolicy module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase import PolicyBase


class DowntimePolicy(PolicyBase):
    """The DowntimePolicy checks for downtimes, scheduled or ongoing, depending on the command parameters."""

    @staticmethod
    def _evaluate(commandResult):
        """It returns Active status if there is no downtime announced.
        Banned if the element is in OUTAGE.
        Degraded if it is on WARNING status.
        Otherwise, it returns error.
        """

        result = {"Status": None, "Reason": None}

        if not commandResult["OK"]:
            result["Status"] = "Error"
            result["Reason"] = commandResult["Message"]
            return S_OK(result)

        status = commandResult["Value"]

        if status is None:
            result["Status"] = "Active"
            result["Reason"] = "No DownTime announced"
            return S_OK(result)

        elif status["Severity"] == "OUTAGE":
            result["Status"] = "Banned"

        elif status["Severity"] == "WARNING":
            result["Status"] = "Degraded"

        else:
            _reason = 'DT_Policy: GOCDB returned an unknown value for DT: "%s"' % status["DowntimeID"]
            result["Status"] = "Error"
            result["Reason"] = _reason
            return S_OK(result)

        # result[ 'EndDate' ] = status[ 'EndDate' ]
        result["Reason"] = "%s %s" % (status["DowntimeID"], status["Description"])
        return S_OK(result)



""" StateMachine

  This module contains the RSS implementation of StateMachine, using its own states map.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.Core.Utilities.StateMachine import State, StateMachine

__RCSID__ = "$Id$"


class RSSMachine(StateMachine):
    """
    RSS implementation of the State Machine. It defines six states, which ordered
    by level conform the following list ( higher level first ): Unknown, Active,
    Degraded, Probing, Banned, Error.

    The StateMachine allows any transition except if the current state is Banned,
    which will force any transition to any state different of Error, Banned and
    Probing to Probing.

      examples:
        >>> rsm0 = RSSMachine( None )
        >>> rsm1 = RSSMachine( 'Unknown' )

    :param state: name of the current state of the StateMachine
    :type state: None or str

    """

    def __init__(self, state):
        """
        Constructor.

        """

        super(RSSMachine, self).__init__(state)

        # Defines state map.
        self.states = {
            "Unknown": State(5),
            "Active": State(4),
            "Degraded": State(3),
            "Probing": State(2),
            "Banned": State(1, ["Error", "Banned", "Probing"], defState="Probing"),
            "Error": State(0),
        }

    def orderPolicyResults(self, policyResults):
        """
        Method built specifically to interact with the policy results obtained on the
        PDP module. It sorts the input based on the level of their statuses, the lower
        the level state, the leftmost position in the list. Beware, if any of the statuses
        is not know to the StateMachine, it will be ordered first, as its level will be
        -1 !.

        examples:
          >>> rsm0.orderPolicyResults( [ { 'Status' : 'Active', 'A' : 'A' },
                                         { 'Status' : 'Banned', 'B' : 'B' } ] )
              [ { 'Status' : 'Banned', 'B' : 'B' }, { 'Status' : 'Active', 'A' : 'A' } ]
          >>> rsm0.orderPolicyResults( [ { 'Status' : 'Active', 'A' : 'A' },
                                         { 'Status' : 'Rubbish', 'R' : 'R' } ] )
              [ { 'Status' : 'Rubbish', 'R' : 'R' }, { 'Status' : 'Active', 'A' : 'A' } ]


        :param policyResults: list of dictionaries to be ordered. The dictionary can have any key as
            far as the key `Status` is present.
        :type policyResults: python:list

        :result: list( dict ), which is ordered
        """

        # We really do not need to return, as the list is mutable
        policyResults.sort(key=self.getLevelOfPolicyState)

    def getLevelOfPolicyState(self, policyResult):
        """
        Returns the level of the state associated with the policy, -1 if something
        goes wrong. It is mostly used while sorting policies with method `orderPolicyResults`.

        examples:
          >>> rsm0.getLevelOfPolicyState( { 'Status' : 'Active', 'A' : 'A' } )
              5
          >>> rsm0.getLevelOfPolicyState( { 'Status' : 'Rubbish', 'R' : 'R' } )
              -1

        :param dict policyResult: dictionary that must have the `Status` key.
        :return: int || -1 ( if policyResult[ 'Status' ] is not known by the StateMachine )
        """

        return self.getLevelOfState(policyResult["Status"])



#! /usr/bin/env python
########################################################################
# File :    dirac-stager-monitor-file
# Author :  Daniela Remenska
########################################################################
"""
Give monitoring information regarding a staging file uniquely identified with (LFN,SE)

- status
- last update
- jobs requesting this file to be staged
- SRM requestID
- pin expiry time
- pin length

Example:
  $ dirac-stager-monitor-file.py /lhcb/LHCb/Collision12/FULL.DST/00020846/0005/0_1.full.dst GRIDKA-RDST
  --------------------
  LFN     : /lhcb/LHCb/Collision12/FULL.DST/00020846/0005/0_1.full.dst
  SE      : GRIDKA-RDST
  PFN     : srm://gridka-dCache.fzk.de/pnfs/gridka.de/lhcb/LHCb/Collision12/FULL.DST/00020846/0005/0_1.full.dst
  Status  : StageSubmitted
  LastUpdate: 2013-06-11 18:13:40
  Reason  : None
  Jobs requesting this file to be staged: 48518896
  ------SRM staging request info--------------
  SRM RequestID: -1768636375
  SRM StageStatus: StageSubmitted
  SRM StageRequestSubmitTime: 2013-06-11 18:13:38
  SRM StageRequestCompletedTime: None
  SRM PinExpiryTime: None
  SRM PinLength: 43200
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("LFN: LFN of the staging file")
    Script.registerArgument("SE: Storage Element for the staging file")
    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import exit as DIRACExit, gLogger

    lfn, se = Script.getPositionalArgs(group=True)

    from DIRAC.StorageManagementSystem.Client.StorageManagerClient import StorageManagerClient

    client = StorageManagerClient()
    res = client.getCacheReplicas({"LFN": lfn, "SE": se})
    if not res["OK"]:
        gLogger.error(res["Message"])
    cacheReplicaInfo = res["Value"]
    if cacheReplicaInfo:
        replicaID = list(cacheReplicaInfo)[0]
        outStr = "\n--------------------"
        outStr += "\n%s: %s" % ("LFN".ljust(8), cacheReplicaInfo[replicaID]["LFN"].ljust(100))
        outStr += "\n%s: %s" % ("SE".ljust(8), cacheReplicaInfo[replicaID]["SE"].ljust(100))
        outStr += "\n%s: %s" % ("PFN".ljust(8), cacheReplicaInfo[replicaID]["PFN"].ljust(100))
        outStr += "\n%s: %s" % ("Status".ljust(8), cacheReplicaInfo[replicaID]["Status"].ljust(100))
        outStr += "\n%s: %s" % ("LastUpdate".ljust(8), str(cacheReplicaInfo[replicaID]["LastUpdate"]).ljust(100))
        outStr += "\n%s: %s" % ("Reason".ljust(8), str(cacheReplicaInfo[replicaID]["Reason"]).ljust(100))

        resTasks = client.getTasks({"ReplicaID": replicaID})

        if resTasks["OK"]:
            # print resTasks['Message']
            outStr += "\nJob IDs requesting this file to be staged:".ljust(8)
            tasks = resTasks["Value"]
            for tid in tasks.keys():
                outStr += " %s " % (tasks[tid]["SourceTaskID"])

        resStageRequests = client.getStageRequests({"ReplicaID": replicaID})

        if not resStageRequests["OK"]:
            gLogger.error(resStageRequests["Message"])

        if resStageRequests["Records"]:
            stageRequests = resStageRequests["Value"]
            outStr += "\n------SRM staging request info--------------"
            for info in stageRequests.values():
                outStr += "\n%s: %s" % ("SRM RequestID".ljust(8), info["RequestID"].ljust(100))
                outStr += "\n%s: %s" % ("SRM StageStatus".ljust(8), info["StageStatus"].ljust(100))
                outStr += "\n%s: %s" % (
                    "SRM StageRequestSubmitTime".ljust(8),
                    str(info["StageRequestSubmitTime"]).ljust(100),
                )
                outStr += "\n%s: %s" % (
                    "SRM StageRequestCompletedTime".ljust(8),
                    str(info["StageRequestCompletedTime"]).ljust(100),
                )
                outStr += "\n%s: %s" % ("SRM PinExpiryTime".ljust(8), str(info["PinExpiryTime"]).ljust(100))
                outStr += "\n%s: %s sec" % ("SRM PinLength".ljust(8), str(info["PinLength"]).ljust(100))
        else:
            outStr += "\nThere are no staging requests submitted to the site yet.".ljust(8)
    else:
        outStr = "\nThere is no such file requested for staging. Check for typo's!"
        # Script.showHelp()
    gLogger.notice(outStr)

    DIRACExit(0)


if __name__ == "__main__":
    main()



""" :mod: ProcessPoolTests
  =======================

  .. module: ProcessPoolTests
  :synopsis: unit tests for ProcessPool
  .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

  unit tests for ProcessPool
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import sys
import random
import time
import threading

import pytest

from DIRAC import gLogger
from DIRAC.Core.Utilities.ProcessPool import ProcessPool

# Mark this entire module as slow
pytestmark = pytest.mark.slow


@pytest.fixture(autouse=True)
def capture_wrap():
    """Avoid https://github.com/pytest-dev/pytest/issues/5502"""
    sys.stderr.close = lambda *args: None
    sys.stdout.close = lambda *args: None
    yield


def ResultCallback(task, taskResult):
    """dummy result callback"""
    print("callback for %s result is %s" % (task.getTaskID(), taskResult))


def ExceptionCallback(task, exec_info):
    """dummy exception callback"""
    print("callback for %s exception is %s" % (task.getTaskID(), exec_info))


def CallableFunc(taskID, timeWait, raiseException=False):
    """global function to be executed in task"""
    print("pid=%s task=%s will sleep for %s s" % (os.getpid(), taskID, timeWait))
    time.sleep(timeWait)
    if raiseException:
        raise Exception("testException")
    return timeWait


class CallableClass(object):
    """callable class to be executed in task"""

    def __init__(self, taskID, timeWait, raiseException=False):
        self.log = gLogger.getSubLogger(self.__class__.__name__ + "/%s" % taskID)
        self.taskID = taskID
        self.timeWait = timeWait
        self.raiseException = raiseException

    def __call__(self):
        self.log.always("pid=%s task=%s will sleep for %s s" % (os.getpid(), self.taskID, self.timeWait))
        time.sleep(self.timeWait)
        if self.raiseException:
            raise Exception("testException")
        return self.timeWait


# global locked lock
gLock = threading.Lock()
# make sure it is locked
gLock.acquire()


# dummy callable locked class
class LockedCallableClass(object):
    """callable and locked class"""

    def __init__(self, taskID, timeWait, raiseException=False):
        self.log = gLogger.getSubLogger(self.__class__.__name__ + "/%s" % taskID)
        self.taskID = taskID
        self.log.always("pid=%s task=%s I'm locked" % (os.getpid(), self.taskID))
        gLock.acquire()
        self.log.always("you can't see that line, object is stuck by gLock")
        self.timeWait = timeWait
        self.raiseException = raiseException
        gLock.release()

    def __call__(self):
        self.log.always("If you see this line, miracle had happened!")
        self.log.always("will sleep for %s" % self.timeWait)
        time.sleep(self.timeWait)
        if self.raiseException:
            raise Exception("testException")
        return self.timeWait


########################################################################
@pytest.fixture()
def processPool():
    gLogger.showHeaders(True)
    log = gLogger.getSubLogger("TaskCallbacksTests")
    processPool = ProcessPool(4, 8, 8)
    processPool.daemonize()
    yield processPool


def test_TaskCallbacks_CallableClass(processPool):
    """CallableClass and task callbacks test"""
    i = 0
    while True:
        if processPool.getFreeSlots() > 0:
            timeWait = random.randint(0, 5)
            raiseException = False
            if not timeWait:
                raiseException = True
            result = processPool.createAndQueueTask(
                CallableClass,
                taskID=i,
                args=(i, timeWait, raiseException),
                callback=ResultCallback,
                exceptionCallback=ExceptionCallback,
                blocking=True,
            )
            if result["OK"]:
                print("CallableClass enqueued to task %s" % i)
                i += 1
            else:
                continue
        if i == 10:
            break
    processPool.finalize(2)


def test_TaskCallbacks_CallableFunc(processPool):
    """CallableFunc and task callbacks test"""
    i = 0
    while True:
        if processPool.getFreeSlots() > 0:
            timeWait = random.randint(0, 5)
            raiseException = False
            if not timeWait:
                raiseException = True
            result = processPool.createAndQueueTask(
                CallableFunc,
                taskID=i,
                args=(i, timeWait, raiseException),
                callback=ResultCallback,
                exceptionCallback=ExceptionCallback,
                blocking=True,
            )
            if result["OK"]:
                print("CallableClass enqueued to task %s" % i)
                i += 1
            else:
                continue
        if i == 10:
            break
    processPool.finalize(2)


########################################################################
@pytest.fixture()
def processPoolWithCallbacks():
    gLogger.showHeaders(True)
    log = gLogger.getSubLogger("ProcessPoolCallbacksTests")
    processPoolWithCallbacks = ProcessPool(
        4,
        8,
        8,
        poolCallback=lambda taskID, taskResult: log.always("callback result for %s is %s" % (taskID, taskResult)),
        poolExceptionCallback=lambda taskID, taskException: log.always(
            "callback exception for %s is %s" % (taskID, taskException)
        ),
    )
    processPoolWithCallbacks.daemonize()
    yield processPoolWithCallbacks


def test_ProcessPoolCallbacks_CallableClass(processPoolWithCallbacks):
    """CallableClass and pool callbacks test"""
    i = 0
    while True:
        if processPoolWithCallbacks.getFreeSlots() > 0:
            timeWait = random.randint(0, 5)
            raiseException = False
            if not timeWait:
                raiseException = True
            result = processPoolWithCallbacks.createAndQueueTask(
                CallableClass,
                taskID=i,
                args=(i, timeWait, raiseException),
                usePoolCallbacks=True,
                blocking=True,
            )
            if result["OK"]:
                print("CallableClass enqueued to task %s" % i)
                i += 1
            else:
                continue
        if i == 10:
            break
    processPoolWithCallbacks.finalize(2)


def test_ProcessPoolCallbacks_CallableFunc(processPoolWithCallbacks):
    """CallableFunc and pool callbacks test"""
    i = 0
    while True:
        if processPoolWithCallbacks.getFreeSlots() > 0:
            timeWait = random.randint(0, 5)
            raiseException = False
            if not timeWait:
                raiseException = True
            result = processPoolWithCallbacks.createAndQueueTask(
                CallableFunc,
                taskID=i,
                args=(i, timeWait, raiseException),
                usePoolCallbacks=True,
                blocking=True,
            )
            if result["OK"]:
                print("CallableFunc enqueued to task %s" % i)
                i += 1
            else:
                continue
        if i == 10:
            break
    processPoolWithCallbacks.finalize(2)


########################################################################
@pytest.fixture()
def processPoolWithCallbacks2():
    gLogger.showHeaders(True)
    log = gLogger.getSubLogger("TaskTimeOutTests")
    processPoolWithCallbacks2 = ProcessPool(
        2,
        4,
        8,
        poolCallback=lambda taskID, taskResult: log.always("callback result for %s is %s" % (taskID, taskResult)),
        poolExceptionCallback=lambda taskID, taskException: log.always(
            "callback exception for %s is %s" % (taskID, taskException)
        ),
    )
    processPoolWithCallbacks2.daemonize()
    yield processPoolWithCallbacks2


def test_TaskTimeOut_CallableClass(processPoolWithCallbacks2):
    """CallableClass and task time out test"""
    i = 0
    while True:
        if processPoolWithCallbacks2.getFreeSlots() > 0:
            timeWait = random.randint(0, 5) * 1
            raiseException = False
            if not timeWait:
                raiseException = True
            result = processPoolWithCallbacks2.createAndQueueTask(
                CallableClass,
                taskID=i,
                args=(i, timeWait, raiseException),
                timeOut=1.5,
                usePoolCallbacks=True,
                blocking=True,
            )
            if result["OK"]:
                print("CallableClass enqueued to task %s timeWait=%s exception=%s" % (i, timeWait, raiseException))
                i += 1
            else:
                continue
        if i == 16:
            break
    processPoolWithCallbacks2.finalize(2)


def test_TaskTimeOut_CallableFunc(processPoolWithCallbacks2):
    """CallableFunc and task timeout test"""
    i = 0
    while True:
        if processPoolWithCallbacks2.getFreeSlots() > 0:
            timeWait = random.randint(0, 5) * 0.5
            raiseException = False
            if not timeWait:
                raiseException = True
            result = processPoolWithCallbacks2.createAndQueueTask(
                CallableFunc,
                taskID=i,
                args=(i, timeWait, raiseException),
                timeOut=1.5,
                usePoolCallbacks=True,
                blocking=True,
            )
            if result["OK"]:
                print("CallableFunc enqueued to task %s timeWait=%s exception=%s" % (i, timeWait, raiseException))
                i += 1
            else:
                continue
        if i == 16:
            break
    processPoolWithCallbacks2.finalize(2)


def test_TaskTimeOut_LockedClass(processPoolWithCallbacks2):
    """LockedCallableClass and task time out test"""
    for loop in range(2):
        print("loop %s" % loop)
        i = 0
        while i < 16:
            if processPoolWithCallbacks2.getFreeSlots() > 0:
                timeWait = random.randint(0, 5) * 0.5
                raiseException = False
                if timeWait == 0.5:
                    raiseException = True
                klass = CallableClass
                if timeWait >= 2.0:
                    klass = LockedCallableClass
                result = processPoolWithCallbacks2.createAndQueueTask(
                    klass,
                    taskID=i,
                    args=(i, timeWait, raiseException),
                    timeOut=1.5,
                    usePoolCallbacks=True,
                    blocking=True,
                )
                if result["OK"]:
                    print(
                        "%s enqueued to task %s timeWait=%s exception=%s"
                        % (klass.__name__, i, timeWait, raiseException)
                    )
                    i += 1
                else:
                    continue
        print("being idle for a while")
        for _ in range(100000):
            for _ in range(1000):
                pass

    print("finalizing...")
    processPoolWithCallbacks2.finalize(10)
    # unlock
    gLock.release()



""" Testing the FCConditionPaserClass
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
from DIRAC.Resources.Catalog.FCConditionParser import FCConditionParser

__RCSID__ = "$Id $"


class TestLogicEvaluation(unittest.TestCase):
    """Tests all the logic evaluation"""

    def setUp(self):
        self.fcp = FCConditionParser()
        self.lfns = ["/lhcb/lfn1", "/lhcb/lfn2"]

    def test_01_simpleParse(self):
        """Test the parse of a single plugin"""

        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=False")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to False
        for lfn in self.lfns:
            self.assertTrue(not res["Value"]["Successful"][lfn], res)

    def test_02_notLogic(self):
        """Testing the ! operator"""

        res = self.fcp("catalogName", "operationName", self.lfns, condition="!Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to False
        for lfn in self.lfns:
            self.assertTrue(not res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="!Dummy=False")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

    def test_03_andLogic(self):
        """Testing the & operator"""

        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=True & Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=False & Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to False
        for lfn in self.lfns:
            self.assertTrue(not res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=True & Dummy=False")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to False
        for lfn in self.lfns:
            self.assertTrue(not res["Value"]["Successful"][lfn], res)

    def test_04_orLogic(self):
        """Testing the | operator"""

        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=True | Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=False | Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=False | Dummy=False")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to False
        for lfn in self.lfns:
            self.assertTrue(not res["Value"]["Successful"][lfn], res)

    def test_05_priority(self):
        """Testing the priority of operators"""

        res = self.fcp("catalogName", "operationName", self.lfns, condition="!Dummy=False & Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="!Dummy=True | Dummy=False")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to False
        for lfn in self.lfns:
            self.assertTrue(not res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=True & Dummy=False | Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=True | Dummy=False & Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="!Dummy=True | Dummy=False & Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to False
        for lfn in self.lfns:
            self.assertTrue(not res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="!Dummy=True | !Dummy=False & Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="!Dummy=True | !Dummy=False & !Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to False
        for lfn in self.lfns:
            self.assertTrue(not res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="[!Dummy=False] & Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="![Dummy=False] & Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="![Dummy=False & Dummy=True]")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="[Dummy=True | Dummy=False] & Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=True | [Dummy=False & Dummy=True]")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=False | [Dummy=False & Dummy=True]")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to False
        for lfn in self.lfns:
            self.assertTrue(not res["Value"]["Successful"][lfn], res)

    def test_06_errors(self):
        """Testing different error situation"""

        # Error in the plugin
        res = self.fcp("catalogName", "operationName", self.lfns, condition="Dummy=CantParse")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to False
        for lfn in self.lfns:
            self.assertTrue(not res["Value"]["Successful"][lfn], res)

        # Non existing plugin
        res = self.fcp("catalogName", "operationName", self.lfns, condition="NonExistingPlugin=something")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to False
        for lfn in self.lfns:
            self.assertTrue(not res["Value"]["Successful"][lfn], res)

        # Error in the grammar
        res = self.fcp("catalogName", "operationName", self.lfns, condition="[Dummy=True")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to False
        for lfn in self.lfns:
            self.assertTrue(not res["Value"]["Successful"][lfn], res)

    def test_07_noCondition(self):
        """Testing different error situation"""

        # Non condition given
        res = self.fcp("catalogName", "operationName", self.lfns, condition="")

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)

        # Can't retrive conditions
        # It so happen that it will all be True
        res = self.fcp("catalogName", "operationName", self.lfns, condition=None)

        self.assertTrue(res["OK"], res)
        # We expect all the lfn to be to True
        for lfn in self.lfns:
            self.assertTrue(res["Value"]["Successful"][lfn], res)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestLogicEvaluation)

    unittest.TextTestRunner(verbosity=2).run(suite)
